{
*****************************************************************************
*  Name    : ARP.bas                                                        *
*  Date    : 28/08/2007                                                     *
*  Version : 1.0 - Based around Microchip TCP/IP stack version 4.13         *
*  Notes   :                                                                *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module ARP

Include "NETTypes.bas"
Include "NETUtils.bas"
Include "MAC.bas"

// ARP Operation codes.
Const 
   ARP_OPERATION_REQ            = $01,
   ARP_OPERATION_RESP           = $02

// ETHERNET packet type as defined by IEEE 802.3
Const
    HW_ETHERNET                 = ($0001),
    ARP_IP                      = ($0800)

// This ARP task caches one ARP response.
#ifdef STACK_CLIENT_MODE
Dim Cache As NODE_INFO
#endif

// ARP packet
Structure ARP_PACKET
   HardwareType As Word
   Protocol As Word
   MACAddrLen As Byte
   ProtocolLen As Byte
   Operation As Word
   SenderMACAddr As MAC_ADDR
   SenderIPAddr As IP_ADDR
   TargetMACAddr As MAC_ADDR
   TargetIPAddr As IP_ADDR
End Structure
   
(*
 *********************************************************************
 * Function:        static void SwapARPPacket(ARP_PACKET* p)
 * PreCondition:    None
 * Input:           p   - ARP packet To be swapped.
 * Output:          None
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)
Sub SwapARPPacket(ByRef P As ARP_PACKET)
   P.HardwareType     = NETUtils.swaps(P.HardwareType)
   P.Protocol         = NETUtils.swaps(P.Protocol)
   P.Operation        = NETUtils.swaps(P.Operation)
End Sub
(*
 *********************************************************************
 * Function:        BOOL ARPPut(NODE_INFO* more, Byte opCode)
 * PreCondition:    None
 * Input:           remote  - Remote node info
 *                  opCode  - ARP op code To send
 * Output:          TRUE - The ARP packet was generated properly
 *					FALSE - Unable To allocate a TX buffer
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)
Function ARPPut(ByRef packet As ARP_PACKET) As Boolean
   While Not MACIsTxReady
   Wend
   MACSetWritePtr(BASE_TX_ADDR)

   packet.HardwareType  = HW_ETHERNET
   packet.Protocol      = ARP_IP
   packet.MACAddrLen    = SizeOf(MAC_ADDR)
   packet.ProtocolLen   = SizeOf(IP_ADDR)
   packet.SenderMACAddr = AppConfig.MyMACAddr	
   packet.SenderIPAddr  = AppConfig.MyIPAddr  
   SwapARPPacket(packet)
   
   MACPutHeader(packet.TargetMACAddr, MAC_ARP, SizeOf(packet))
   MACPutArray(@packet, SizeOf(ARP_PACKET))
   
   MACFlush()
   result = true   
End Function
(*
 *********************************************************************
 * Function:        void ARPInit(void)
 * PreCondition:    None
 * Input:           None
 * Output:          ARP Cache is initialized.
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)

Public Sub ARPInit()
#ifdef STACK_CLIENT_MODE
   Cache.MACAddr.v(0) = $ff
   Cache.MACAddr.v(1) = $ff
   Cache.MACAddr.v(2) = $ff
   Cache.MACAddr.v(3) = $ff
   Cache.MACAddr.v(4) = $ff
   Cache.MACAddr.v(5) = $ff
   Cache.IPAddr.Val = $0
#endif
End Sub
   
// enumerations...
Const
   SM_ARP_IDLE   = 0,
   SM_ARP_REPLY  = 1
   
// statics for ARPProcess()   
Dim 
   static_smARP As Byte,
   static_Target As NODE_INFO
   
(*
 *********************************************************************
 * Function:        BOOL ARPProcess(void)
 * PreCondition:    ARP packet is ready in MAC buffer.
 * Input:           None
 * Output:          TRUE: If all processing of this ARP packet is complete.  
 *                  Do Not call ARPProcess() again Until a new ARP packet is 
 *                  waiting in the RX buffer.
 *					FALSE: If ARPProcess() must be called more times.  We 
 *                  need more time To sent an ARP response
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)
Public Function ARPProcess() As Boolean
   Dim packet As ARP_PACKET
   Dim Target As static_Target
   Dim smARP As static_smARP

   If smARP = SM_ARP_IDLE Then
      
      // Obtain the incoming ARP packet
      MACGetArray(@packet, SizeOf(packet))		
      MACDiscardRx()
      SwapARPPacket(packet)
	
      // Validate the ARP packet
      If packet.HardwareType <> HW_ETHERNET Or packet.MACAddrLen <> SizeOf(MAC_ADDR) Or packet.ProtocolLen <> SizeOf(IP_ADDR) Then
         result = true
         Exit
      EndIf
		
      // Handle incoming ARP responses
      #ifdef STACK_CLIENT_MODE
      If packet.Operation = ARP_OPERATION_RESP Then
         Cache.MACAddr = packet.SenderMACAddr
         Cache.IPAddr = packet.SenderIPAddr
         result = true
         Exit
      EndIf
      #endif
        
      // Handle incoming ARP requests for our MAC address
      If packet.Operation = ARP_OPERATION_REQ Then
	     If packet.TargetIPAddr.Val <> AppConfig.MyIPAddr.Val Then
            result = true
            Exit
         EndIf
         Target.IPAddr = packet.SenderIPAddr
         Target.MACAddr = packet.SenderMACAddr
         smARP = SM_ARP_REPLY // Do not break.  If we get down here, we need to send a reply.
      EndIf
   EndIf
   
   If smARP = SM_ARP_REPLY Then
      packet.Operation = ARP_OPERATION_RESP
      packet.TargetMACAddr = Target.MACAddr
      packet.TargetIPAddr = Target.IPAddr
      
      // Send an ARP response to a previously received request
      If Not ARPPut(packet) Then
         result = false
         Exit
      EndIf

      // Begin listening for ARP requests again
      smARP = SM_ARP_IDLE
   EndIf
      
   result = true
End Function
(*
 *********************************************************************
 * Function:        void ARPResolve(IP_ADDR* IPAddr)
 * PreCondition:    MACIsTxReady(TRUE) returns TRUE
 * Input:           IPAddr  - IP Address To be resolved.
 * Output:          None
 * Side Effects:    None
 * Overview:        An ARP request is sent.
 * Note:            This Function is available only when
 *                  STACK_CLIENT_MODE is defined.
 ********************************************************************/
*)
Public Sub ARPResolve(ByRef IPAddr As IP_ADDR)
   Dim packet As ARP_PACKET

   packet.Operation            = ARP_OPERATION_REQ
   packet.TargetMACAddr.v(0)   = $ff
   packet.TargetMACAddr.v(1)   = $ff
   packet.TargetMACAddr.v(2)   = $ff
   packet.TargetMACAddr.v(3)   = $ff
   packet.TargetMACAddr.v(4)   = $ff
   packet.TargetMACAddr.v(5)   = $ff

   // ARP query either the IP address directly (on our subnet), or do an ARP query for our Gateway if off of our subnet
   If ((AppConfig.MyIPAddr.Val Xor IPAddr.Val) And AppConfig.MyMask.Val) <> 0 Then
      packet.TargetIPAddr = AppConfig.MyGateway
   Else
      packet.TargetIPAddr = IPAddr
   EndIf
          
   ARPPut(packet)
End Sub
(*
 *********************************************************************
 * Function:        BOOL ARPIsResolved(IP_ADDR* IPAddr,
 *                                      MAC_ADDR *MACAddr)
 * PreCondition:    None
 * Input:           IPAddr      - IPAddress To be resolved.
 *                  MACAddr     - Buffer To hold corresponding
 *                                MAC Address.
 * Output:          TRUE If given IP Address has been resolved.
 *                  FALSE otherwise.
 * Side Effects:    None
 * Overview:        None
 * Note:            This Function is available only when
 *                  STACK_CLIENT_MODE is defined.
 ********************************************************************/
*)
Public Function ARPIsResolved(ByRef IPAddr As IP_ADDR, ByRef MACAddr As MAC_ADDR) As Boolean
   #ifdef STACK_CLIENT_MODE
   If (Cache.IPAddr.Val = IPAddr.Val) Or 
	  ((Cache.IPAddr.Val = AppConfig.MyGateway.Val) And ((AppConfig.MyIPAddr.Val Xor IPAddr.Val) And AppConfig.MyMask.Val) <> 0) Then
      MACAddr = Cache.MACAddr
      result = true
      Exit
   EndIf
   #endif
   result = false       
End Function

// initailise the module
static_smARP = SM_ARP_IDLE  // used in ARPProcess()


