{
*****************************************************************************
*  Name    : DNS.BAS                                                        *
*  Date    : 07/09/2007                                                     *
*  Version : 1.0 - Based around Microchip TCP/IP stack version 4.13         *
*  Notes   :                                                                *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module DNS

Include "NETConfig.bas"
Include "NETTypes.bas"
Include "NETUtils.bas"
Include "ARP.bas"
Include "UDP.bas"
Include "Tick.bas"

// Type constants for DNSResolve(HostName, Type)
Public Const
   DNS_TYPE_A       = (1),
   DNS_TYPE_MX      = (15)

Const
   DNS_PORT         = 53,
   DNS_TIMEOUT      = (TICK_SECOND * 1),
   Nil              = 0

Const
   DNS_ARP_START_RESOLVE = 0,
   DNS_ARP_RESOLVE = 1,
   DNS_ARP_START_RESOLVE2 = 2,
   DNS_ARP_RESOLVE2 = 3,
   DNS_ARP_START_RESOLVE3 = 4,
   DNS_ARP_RESOLVE3 = 5,
   DNS_ARP_FAIL = 6,
   DNS_OPEN_SOCKET = 7,
   DNS_QUERY = 8,
   DNS_GET_RESULT = 9,
   DNS_QUERY2 = 10,
   DNS_GET_RESULT2 = 11,
   DNS_QUERY3 = 12,
   DNS_GET_RESULT3 = 13,
   DNS_QUERY_FAIL = 14,
   DNS_DONE = 15

Structure TFlagBits
   DNSInUse As Boolean
   AddressValid As Boolean
End Structure

Structure TFlags
   Val As Byte
   Bits As TFlagBits Union
End Structure

Dim
   MySocket As UDP_SOCKET,
   DNSHostName As Word,
   DNSHostNameROM As TROMAddress,
   RecordType As Byte,
   ResolvedInfo As NODE_INFO,
   Flags As TFlags,
   smDNS As Byte,
   DNSServer As IP_ADDR

Structure DNS_HEADER
   TransactionID As WORD_VAL
   Flags As WORD_VAL
   Questions As WORD_VAL
   Answers As WORD_VAL
   AuthoritativeRecords As WORD_VAL
   AdditionalRecords As WORD_VAL
End Structure

Structure DNS_ANSWER_HEADER
   ResponseName As WORD_VAL
   ResponseType As WORD_VAL
   ResponseClass As WORD_VAL
   ResponseTTL As DWORD_VAL
   ResponseLen As WORD_VAL
End Structure

(*
 *********************************************************************
 * Function:        BOOL DNSBeginUsage(void)
 * PreCondition:    Stack is initialized()
 * Input:           None
 * Output:          TRUE: If no DNS resolution operations are in progress And this 
 *                  application has successfully taken ownership of the DNS Module
 *					FALSE: If the DNS Module is currently being used by some other 
 *                  Module.  Call DNSBeginUsage() some time later (after returning To 
 *                  the main() Program loop).
 * Side Effects:    None
 * Overview:        Call DNSBeginUsage() And make sure it returns TRUE before 
 *                  calling any DNS APIs.  Call DNSEndUsage() when this application 
 *                  no longer needs the DNS Module so that other applications may 
 *                  make use of it.
 * Note:            None
 ********************************************************************/
*)
Public Function DNSBeginUsage() As Boolean
   If Flags.Bits.DNSInUse Then
      result = false
      Exit
   EndIf
   DNSServer = AppConfig.PrimaryDNSServer
   Flags.Bits.DNSInUse = true
   result = true
End Function
(*
 *********************************************************************
 * Function:        BOOL DNSEndUsage(void)
 * PreCondition:    DNSBeginUsage() returned TRUE on a previous call.
 * Input:           None
 * Output:          TRUE: If the address To the host name was 
 *						  successfully resolved
 *					FALSE: If the DNS failed Or address does Not 
 *						   exists.  The contents of *Hostname are 
 *						   unchanged.
 * Side Effects:    None
 *
 * Overview:        Call DNSBeginUsage() And make sure it returns 
 *					TRUE before calling any DNS APIs.  Call 
 *					DNSEndUsage() when this application no longer 
 *					needs the DNS Module so that other applications 
 *					may make use of it.
 * Note:            None
 ********************************************************************/
*)
Public Function DNSEndUsage() As Boolean
   If MySocket <> INVALID_UDP_SOCKET Then
      UDPClose(MySocket)
      MySocket = INVALID_UDP_SOCKET
   EndIf   
   smDNS = DNS_DONE
   Flags.Bits.DNSInUse = false
   result = Flags.Bits.AddressValid
End Function
(*
 *********************************************************************
 * Function:        void DNSResolve(Byte *Hostname, Byte RecordType)
 * PreCondition:    Stack is initialized()
 * Input:           *Hostname: Null terminated String specifying the 
 *							   host address To resolve To an IP 
 *							   address.
 *					Type: DNS_TYPE_A (1d): Host Address (normal IP address)
 *						  DNS_TYPE_MX (15d): Mail eXchange
 *						  All other values are reserved
 * Output:          None
 * Side Effects:    None
 * Overview:        Call DNSIsResolved() Until the host is resolved.
 * Note:            A UDP Socket must be available before this 
 *					Function is called.  It is freed at the End of 
 *					the resolution.  MAX_UDP_SOCKETS may need To be 
 *					increased If other modules use UDP sockets.
 *					You must Not modify *Hostname Until DNSIsResolved() 
 *					returns TRUE.
 ********************************************************************/
*)
Public Sub DNSResolve(Hostname As Word, _Type As Byte)
   If StringToIPAddress(Hostname, ResolvedInfo.IPAddr) Then
      Flags.Bits.AddressValid = true
      smDNS = DNS_DONE
   Else
      DNSHostName = Hostname
	  DNSHostNameROM = Nil
	  smDNS = DNS_ARP_START_RESOLVE
	  RecordType = _Type
	  Flags.Bits.AddressValid = false
   EndIf
End Sub
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Public Sub DNSResolveROM(Hostname As TROMAddress, _Type As Byte)
   If ROMStringToIPAddress(Hostname, ResolvedInfo.IPAddr) Then
      Flags.Bits.AddressValid = true
      smDNS = DNS_DONE
   Else
      DNSHostName = Nil
      DNSHostNameROM = Hostname
      smDNS = DNS_ARP_START_RESOLVE
      RecordType = _Type
	  Flags.Bits.AddressValid = false
   EndIf
End Sub
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Sub DNSPutString(Str As Word)
   Dim RightPtr As Word
   Dim i As Byte
   Dim Len As Byte

   RightPtr = Str
   While true
      Repeat
         FSR0 = RightPtr
         Inc(RightPtr)
         i = INDF0
      Until i = $00 Or i = Byte(".") Or i = Byte("/") Or i = Byte(",") Or i = Byte(">") 
	
	  // Put the length and data
	  // Also, skip over the '.' in the input string	
	  Len = Byte(RightPtr - Str) - 1
	  UDPPut(Len)
      Str = Str + UDPPutArray(Str, Len) + 1

	  If i = $00 Or i = Byte("/") Or i = Byte(",") Or i = Byte(">") Then
	     Break
	  EndIf
   Wend
	
   // Put the string terminator character
   UDPPut($00)
End Sub
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Sub DNSPutROMString(Str As TROMAddress)
   Dim RightPtr As Word
   Dim i As Byte
   Dim Len As Byte

   RightPtr = Str
   While true
      Repeat
         i = ReadROM(RightPtr)
         Inc(RightPtr)
      Until i = $00 Or i = Byte(".") Or i = Byte("/") Or i = Byte(",") Or i = Byte(">") 
	
	  // Put the length and data
	  // Also, skip over the '.' in the input string	
	  Len = Byte(RightPtr - Str) - 1
	  UDPPut(Len)
      Str = Str + UDPPutROMArray(Str, Len) + 1

	  If i = $00 Or i = Byte("/") Or i = Byte(",") Or i = Byte(">") Then
	     Break
	  EndIf
   Wend
	
   // Put the string terminator character
   UDPPut($00)
End Sub
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Sub DNSGetString(Str As Word)
   Dim i As Byte
   While true
      If Not UDPGet(i) Then
         Exit
      EndIf
      If i = 0 Then
         Exit
      EndIf
      i = UDPGetArray(Str, i)
      If Str <> 0 Then
         Str = Str + i
      EndIf
   Wend
End Sub
(*
/*********************************************************************
 * Function:        BOOL DNSIsResolved(IP_ADDR *HostIP)
 * PreCondition:    DNSResolve() was called.
 * Input:           HostIP: Pointer to IP_ADDR structure to store the 
 *							returned host IP address when DNS 
 *							resolution is complete.
 * Output:          BOOL: TRUE if the DNS client has gotten the IP 
 *						  address or timed out
 *						  FALSE if the DNS client is still working.
 *					*HostIP: 4 byte IP address. 0.0.0.0 on timeout.
 * Side Effects:    None
 * Overview:        Call DNSIsResolved() until the host is resolved.
 * Note:            You cannot start two DNS resolution proceedures 
 *					concurrently.
 *					You must not modify *Hostname until DNSIsResolved() 
 *					returns TRUE.
 ********************************************************************/
*)
// statics...
Dim
   StartTime As TICK,
   SentTransactionID As WORD_VAL

// function...   
Public Function DNSIsResolved(ByRef HostIP As IP_ADDR) As Boolean
   Dim i As Byte
   Dim w As WORD_VAL
   Dim DNSHeader As DNS_HEADER
   Dim DNSAnswerHeader As DNS_ANSWER_HEADER

   If smDNS = DNS_ARP_START_RESOLVE Or smDNS = DNS_ARP_START_RESOLVE2 Or smDNS = DNS_ARP_START_RESOLVE3 Then
      ARPResolve(DNSServer)
      StartTime = TickGet()
      Inc(smDNS)
      GoTo DoBreak
   EndIf

   If smDNS = DNS_ARP_RESOLVE Or smDNS = DNS_ARP_RESOLVE2 Or smDNS = DNS_ARP_RESOLVE3 Then
      If Not ARPIsResolved(DNSServer, ResolvedInfo.MACAddr) Then
         If TickGet() - StartTime > DNS_TIMEOUT Then
            Inc(smDNS)
         EndIf
         GoTo DoBreak
      EndIf
      ResolvedInfo.IPAddr.Val = DNSServer.Val
      smDNS = DNS_OPEN_SOCKET
   EndIf   
   // No break: DNS_OPEN_SOCKET is the correct next state
		
   If smDNS = DNS_OPEN_SOCKET Then
      MySocket = UDPOpen(0, ResolvedInfo, DNS_PORT)
      If MySocket = INVALID_UDP_SOCKET Then
         GoTo DoBreak
      EndIf     
      smDNS = DNS_QUERY
      // No need to break, we can immediately start resolution
   EndIf
      
   If smDNS = DNS_QUERY Or smDNS = DNS_QUERY2 Or smDNS = DNS_QUERY3 Then
      If UDPIsPutReady(MySocket) = 0 Then
         GoTo DoBreak
      EndIf

      // Put DNS query here
      Inc(SentTransactionID.Val)
	  UDPPut(SentTransactionID.v(1))// User chosen transaction ID
	  UDPPut(SentTransactionID.v(0))
	  UDPPut($01)		// Standard query with recursion
	  UDPPut($00)	
	  UDPPut($00)		// 0x0001 questions
	  UDPPut($01)
	  UDPPut($00)		// 0x0000 answers
	  UDPPut($00)
	  UDPPut($00)		// 0x0000 name server resource records
	  UDPPut($00)
	  UDPPut($00)		// 0x0000 additional records
	  UDPPut($00)

	  // Put hostname string to resolve
	  If DNSHostName <> 0 Then
         DNSPutString(DNSHostName)  
      Else    
	     DNSPutROMString(DNSHostNameROM)
	  EndIf   

	  UDPPut($00)		// Type: DNS_TYPE_A (host address) or DNS_TYPE_MX for mail exchange
	  UDPPut(RecordType)
	  UDPPut($00)		// Class: IN (Internet)
	  UDPPut($01)

	  UDPFlush()
	  StartTime = TickGet()
	  Inc(smDNS)
	  GoTo DoBreak
   EndIf
   
   If smDNS = DNS_GET_RESULT Or smDNS = DNS_GET_RESULT2 Or smDNS = DNS_GET_RESULT3 Then

      If UDPIsGetReady(MySocket) = 0 Then
	     If TickGet() - StartTime > DNS_TIMEOUT Then
	        Inc(smDNS)
	     EndIf	     
         GoTo DoBreak
      EndIf

      // Retrieve the DNS header and de-big-endian it
	  UDPGet(DNSHeader.TransactionID.v(1))
	  UDPGet(DNSHeader.TransactionID.v(0))

	  // Throw this packet away if it isn't in response to our last query
	  If DNSHeader.TransactionID.Val <> SentTransactionID.Val Then
 	     UDPDiscard()
		 GoTo DoBreak
      EndIf
  
      UDPGet(DNSHeader.Flags.v(1))
	  UDPGet(DNSHeader.Flags.v(0))
	  UDPGet(DNSHeader.Questions.v(1))
	  UDPGet(DNSHeader.Questions.v(0))
	  UDPGet(DNSHeader.Answers.v(1))
	  UDPGet(DNSHeader.Answers.v(0))
	  UDPGet(DNSHeader.AuthoritativeRecords.v(1))
	  UDPGet(DNSHeader.AuthoritativeRecords.v(0))
	  UDPGet(DNSHeader.AdditionalRecords.v(1))
	  UDPGet(DNSHeader.AdditionalRecords.v(0))

      // Remove all questions
	  While DNSHeader.Questions.Val <> 0
         DNSGetString(Nil)
         UDPGet(w.v(1))		// Question type
		 UDPGet(w.v(0))
		 UDPGet(w.v(1))		// Question class
		 UDPGet(w.v(0))
		 DNSHeader.Questions.Val = DNSHeader.Questions.Val - 1
      Wend
			
      // Scan through answers
	  While DNSHeader.Answers.Val <> 0
  	     UDPGet(DNSAnswerHeader.ResponseName.v(1))		// Response name
		 UDPGet(DNSAnswerHeader.ResponseName.v(0))
		 UDPGet(DNSAnswerHeader.ResponseType.v(1))		// Response type
		 UDPGet(DNSAnswerHeader.ResponseType.v(0))
		 UDPGet(DNSAnswerHeader.ResponseClass.v(1))	// Response class
		 UDPGet(DNSAnswerHeader.ResponseClass.v(0))
		 UDPGet(DNSAnswerHeader.ResponseTTL.v(3))		// Time to live
		 UDPGet(DNSAnswerHeader.ResponseTTL.v(2))
		 UDPGet(DNSAnswerHeader.ResponseTTL.v(1))
		 UDPGet(DNSAnswerHeader.ResponseTTL.v(0))
		 UDPGet(DNSAnswerHeader.ResponseLen.v(1))		// Response length
		 UDPGet(DNSAnswerHeader.ResponseLen.v(0))

         // Make sure that this is a 4 byte IP address, response type A or MX, class 1
		 // Check if this is Type A or MX
		 If DNSAnswerHeader.ResponseType.Val = $0001 And 
            DNSAnswerHeader.ResponseClass.Val = $0001 And 
			DNSAnswerHeader.ResponseLen.Val = $0004 Then
			   Flags.Bits.AddressValid = true
			   UDPGet(ResolvedInfo.IPAddr.v(0))
			   UDPGet(ResolvedInfo.IPAddr.v(1))
			   UDPGet(ResolvedInfo.IPAddr.v(2))
			   UDPGet(ResolvedInfo.IPAddr.v(3))
			   GoTo DoneSearchingRecords
		 Else
		    While DNSAnswerHeader.ResponseLen.Val <> 0
			   UDPGet(i)
			   DNSAnswerHeader.ResponseLen.Val = DNSAnswerHeader.ResponseLen.Val - 1
			Wend
         EndIf
         DNSHeader.Answers.Val = DNSHeader.Answers.Val - 1
      Wend

      // Remove all Authoritative Records
	  While DNSHeader.AuthoritativeRecords.Val <> 0
	     UDPGet(DNSAnswerHeader.ResponseName.v(1))		// Response name
		 UDPGet(DNSAnswerHeader.ResponseName.v(0))
		 UDPGet(DNSAnswerHeader.ResponseType.v(1))		// Response type
		 UDPGet(DNSAnswerHeader.ResponseType.v(0))
		 UDPGet(DNSAnswerHeader.ResponseClass.v(1))	// Response class
		 UDPGet(DNSAnswerHeader.ResponseClass.v(0))
		 UDPGet(DNSAnswerHeader.ResponseTTL.v(3))		// Time to live
		 UDPGet(DNSAnswerHeader.ResponseTTL.v(2))
		 UDPGet(DNSAnswerHeader.ResponseTTL.v(1))
		 UDPGet(DNSAnswerHeader.ResponseTTL.v(0))
		 UDPGet(DNSAnswerHeader.ResponseLen.v(1))		// Response length
		 UDPGet(DNSAnswerHeader.ResponseLen.v(0))

		 // Make sure that this is a 4 byte IP address, response type A or MX, class 1
		 // Check if this is Type A
		 If  DNSAnswerHeader.ResponseType.Val = $0001 And
		     DNSAnswerHeader.ResponseClass.Val = $0001 And  // Internet class
			 DNSAnswerHeader.ResponseLen.Val = $0004 Then
			    Flags.Bits.AddressValid = true
				UDPGet(ResolvedInfo.IPAddr.v(0))
				UDPGet(ResolvedInfo.IPAddr.v(1))
				UDPGet(ResolvedInfo.IPAddr.v(2))
				UDPGet(ResolvedInfo.IPAddr.v(3))
				GoTo DoneSearchingRecords
		 Else
		    While DNSAnswerHeader.ResponseLen.Val <> 0
			   UDPGet(i)
			   DNSAnswerHeader.ResponseLen.Val = DNSAnswerHeader.ResponseLen.Val - 1
			Wend
         EndIf
         DNSHeader.AuthoritativeRecords.Val = DNSHeader.AuthoritativeRecords.Val - 1
      Wend

      // Remove all Additional Records
	  While DNSHeader.AdditionalRecords.Val <> 0
         UDPGet(DNSAnswerHeader.ResponseName.v(1))		// Response name
		 UDPGet(DNSAnswerHeader.ResponseName.v(0))
		 UDPGet(DNSAnswerHeader.ResponseType.v(1))		// Response type
		 UDPGet(DNSAnswerHeader.ResponseType.v(0))
		 UDPGet(DNSAnswerHeader.ResponseClass.v(1))	// Response class
		 UDPGet(DNSAnswerHeader.ResponseClass.v(0))
		 UDPGet(DNSAnswerHeader.ResponseTTL.v(3))		// Time to live
		 UDPGet(DNSAnswerHeader.ResponseTTL.v(2))
		 UDPGet(DNSAnswerHeader.ResponseTTL.v(1))
		 UDPGet(DNSAnswerHeader.ResponseTTL.v(0))
		 UDPGet(DNSAnswerHeader.ResponseLen.v(1))		// Response length
		 UDPGet(DNSAnswerHeader.ResponseLen.v(0))

         // Make sure that this is a 4 byte IP address, response type A or MX, class 1
		 // Check if this is Type A
		 If DNSAnswerHeader.ResponseType.Val = $0001 And
			DNSAnswerHeader.ResponseClass.Val = $0001 And // Internet class
			DNSAnswerHeader.ResponseLen.Val = $0004 Then
			   Flags.Bits.AddressValid = true
			   UDPGet(ResolvedInfo.IPAddr.v(0))
			   UDPGet(ResolvedInfo.IPAddr.v(1))
			   UDPGet(ResolvedInfo.IPAddr.v(2))
			   UDPGet(ResolvedInfo.IPAddr.v(3))
			   GoTo DoneSearchingRecords
		 Else
		    While DNSAnswerHeader.ResponseLen.Val <> 0 
			   UDPGet(i)
			   DNSAnswerHeader.ResponseLen.Val = DNSAnswerHeader.ResponseLen.Val - 1
			Wend
		 EndIf
         DNSHeader.AdditionalRecords.Val = DNSHeader.AdditionalRecords.Val - 1
      Wend
            
      DoneSearchingRecords:
         UDPDiscard()
		 UDPClose(MySocket)
		 MySocket = INVALID_UDP_SOCKET
		 smDNS = DNS_DONE
		 // No break, DNS_DONE is the correct step
   EndIf

   If smDNS = DNS_DONE Then
      If Flags.Bits.AddressValid Then
         HostIP.Val = ResolvedInfo.IPAddr.Val
      Else
         HostIP.Val = $00000000
      EndIf      
      result = true
      Exit
   EndIf
      
   If smDNS = DNS_ARP_FAIL Or smDNS = DNS_QUERY_FAIL Then
      
      // DJB - switch to secondary DNS server...
      If DNSServer.Val = AppConfig.PrimaryDNSServer.Val And AppConfig.PrimaryDNSServer.Val <> AppConfig.SecondaryDNSServer.Val Then
         If MySocket <> INVALID_UDP_SOCKET Then
            UDPClose(MySocket)
         EndIf
         DNSServer = AppConfig.SecondaryDNSServer
         smDNS = DNS_ARP_START_RESOLVE
         GoTo DoBreak
      EndIf
      
      HostIP.Val = $00000000
      result = true
      Exit
   EndIf
   
DoBreak:
   result = false
End Function


// initialise
#if defined(STACK_USE_DNS)
MySocket = INVALID_UDP_SOCKET
Flags.Val = 0
smDNS = DNS_DONE
#endif


