{
*****************************************************************************
*  Name    : ICMP.BAS                                                       *
*  Date    : 02/09/2007                                                     *
*  Version : 1.0 - Based around Microchip TCP/IP stack version 4.13         *
*  Notes   :                                                                *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module ICMP

Include "NETConfig.bas"
Include "NETTypes.bas"
Include "NETUtils.bas"
Include "MAC.bas"
Include "IP.bas"
Include "ARP.bas"
Include "Tick.bas"

Const
   ICMP_TIMEOUT	    = (4 * TICK_SECOND)

Structure ICMP_HEADER
   vType As Byte
   vCode As Byte
   wvChecksum As WORD_VAL
   wvIdentifier As WORD_VAL
   wvSequenceNumber As WORD_VAL
End Structure

Structure TICMPFlags
   Val As Byte
   bICMPInUse As Val.Booleans(0)
   bReplyValid As Val.Booleans(1)
End Structure

Dim 
   ICMPHeader As ICMP_HEADER,
   ICMPTimer As TICK,
   ICMPRemote As NODE_INFO,
   ICMPFlags As TICMPFlags
   
Type TICMPState = Byte
Const
   SM_IDLE          = 0,
   SM_ARP_RESOLVE   = 1,
   SM_GET_ECHO      = 2
Dim ICMPState As TICMPState   
(*
 *********************************************************************
 * Function:        void ICMPProcess(void)
 * PreCondition:    MAC buffer contains ICMP type packet.
 * Input:           *remote: Pointer to a NODE_INFO structure of the 
 *					ping requester
 *					len: Count of how many bytes the ping header and 
 *					payload are in this IP packet
 * Output:          Generates an echo reply, if requested
 *					Validates and sets ICMPFlags.bReplyValid if a 
 *					correct ping response to one of ours is received.
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)
Public Sub ICMPProcess(ByRef remote As NODE_INFO, Len As Word)
   Dim dwVal As DWORD_VAL

   // Obtain the ICMP header Type, Code, and Checksum fields
   MACGetArray(@dwVal, SizeOf(dwVal))

   // See if this is an ICMP echo (ping) request
   If dwVal.w(0) = $0008 Then

      // Validate the checksum using the Microchip MAC's DMA module
	  // The checksum data includes the precomputed checksum in the 
	  // header, so a valid packet will always have a checksum of 
	  // 0x0000 if the packet is not disturbed.
	  If MACCalcRxChecksum(0 + SizeOf(IP_HEADER), Len) <> 0 Then
	     Exit
	  EndIf

      // Calculate new Type, Code, and Checksum values
	  dwVal.v(0) = $00	// Type: 0 (ICMP echo/ping reply)
	  dwVal.v(2) = dwVal.v(2) + 8	// Subtract 0x0800 from the checksum
	  If dwVal.v(2) < 8 Then
	     dwVal.v(3) = dwVal.v(3) + 1
		 If dwVal.v(3) = 0 Then
		    dwVal.v(2) = dwVal.v(2) + 1
		 EndIf   
	  EndIf

      // Position the write pointer for the next IPPutHeader operation
	  MACSetWritePtr(BASE_TX_ADDR + SizeOf(ETHER_HEADER))
	
      // Wait for TX hardware to become available (finish transmitting 
	  // any previous packet)
	  While Not IPIsTxReady 
      Wend
	
      // Create IP header in TX memory
	  IPPutHeader(remote, IP_PROT_ICMP, Len)
	  
      // Copy ICMP response into the TX memory
	  MACPutArray(@dwVal, SizeOf(dwVal))
	  MACMemCopyAsync(Word(-1), Word(-1), Len - 4)
	  While Not MACIsMemCopyDone()
      Wend
      
      // Transmit the echo reply packet
	  MACFlush()
   ElseIf dwVal.w(0) = $0000 Then	// See if this an ICMP Echo reply to our request

      // Get the sequence number and identifier fields
	  MACGetArray(@dwVal, SizeOf(dwVal))

      // See if the identifier matches the one we sent
	  If dwVal.w(0) <> $EFBE Then
         Exit
      EndIf

      If dwVal.w(1) <> ICMPHeader.wvSequenceNumber.Val Then
         Exit
      EndIf

      // Validate the ICMP checksum field
	  IPSetRxBuffer(0)
	  If CalcIPBufferChecksum(SizeOf(ICMP_HEADER)) <> 0 Then
	     Exit
	  EndIf
		
      // Flag that we received the response and stop the timer ticking
	  ICMPFlags.bReplyValid = true
      ICMPTimer = TickGet() - ICMPTimer
   EndIf
End Sub
(*
 *********************************************************************
 * Function:        void ICMPSendPing(DWORD dwRemoteIP)
 * PreCondition:    ICMPBeginUsage() returned TRUE
 * Input:           dwRemoteIP: IP Address to ping.  Must be stored 
 *								big endian.  Ex: 192.168.0.1 should be
 *								passed as 0xC0A80001.
 * Output:          Begins the process of transmitting an ICMP echo 
 *					request.  This normally involves an ARP 
 *					resolution procedure first.
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)
Public Sub ICMPSendPing(dwRemoteIP As DWORD)
   
   // Figure out the MAC address to send to
   ICMPRemote.IPAddr.Val = dwRemoteIP
   ARPResolve(ICMPRemote.IPAddr)

   // Set up the ping packet
   ICMPHeader.vType = $08	// 0x08: Echo (ping) request
   ICMPHeader.vCode = $00
   ICMPHeader.wvChecksum.Val = $0000
   ICMPHeader.wvIdentifier.Val = $EFBE
   ICMPHeader.wvSequenceNumber.Val = ICMPHeader.wvSequenceNumber.Val + 1
   ICMPHeader.wvChecksum.Val = NETUtils.CalcIPChecksum(@ICMPHeader, SizeOf(ICMPHeader))

   // Kick off the ICMPGetReply() state machine
   ICMPTimer = TickGet()
   ICMPFlags.bReplyValid = false
   ICMPState = SM_ARP_RESOLVE
End Sub
(*
 *********************************************************************
 * Function:        LONG ICMPGetReply(void)
 * PreCondition:    ICMPBeginUsage() returned TRUE and ICMPSendPing() 
 *					was called
 * Input:           None
 * Output:          -2: No response received yet
 *					-1: Operation timed out (longer than ICMP_TIMEOUT) 
 *						has elapsed.
 *					>=0: Number of TICKs that elapsed between 
 *						 initial ICMP transmission and reception of 
 *						 a valid echo.
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)
Public Function ICMPGetReply() As LONG
   Select ICMPState
      Case SM_IDLE:
	     result = -1
         Exit
      
      Case SM_ARP_RESOLVE:

         // See if the ARP reponse was successfully received
		 If ARPIsResolved(ICMPRemote.IPAddr, ICMPRemote.MACAddr) Then

		    // Position the write pointer for the next IPPutHeader operation
			MACSetWritePtr(BASE_TX_ADDR + SizeOf(ETHER_HEADER))
			
            // Wait for TX hardware to become available (finish transmitting 
			// any previous packet)
			While Not IPIsTxReady
            Wend

			// Create IP header in TX memory
			IPPutHeader(ICMPRemote, IP_PROT_ICMP, SizeOf(ICMP_HEADER))
			MACPutArray(@ICMPHeader, SizeOf(ICMPHeader))
			MACFlush()

            // MAC Address resolved and echo sent, advance state
            ICMPState = SM_GET_ECHO
			result = -2		
            Exit
         EndIf

         // See if the ARP/echo request timed out
		 If TickGet - ICMPTimer > ICMP_TIMEOUT Then
            ICMPState = SM_IDLE
			result = -1
            Exit
         EndIf
         
         result = -2
         Exit
      Case SM_GET_ECHO:
	     // See if the echo was successfully received
		 If ICMPFlags.bReplyValid Then
            result = ICMPTimer
            Exit
         EndIf

         // See if the ARP/echo request timed out
         If TickGet - ICMPTimer > ICMP_TIMEOUT Then
		    ICMPState = SM_IDLE
		    result = -1
		    Exit
		 EndIf
		
		// No echo response back yet
		result = -2
   End Select
End Function
(*
 *********************************************************************
 * Function:        BOOL ICMPBeginUsage(void)
 * PreCondition:    None
 * Input:           None
 * Output:          TRUE: You have successfully gained ownership of 
 *						  the ICMP client module and can now use the 
 *						  ICMPSendPing() and ICMPGetReply() functions.
 *					FALSE: Some other application is using the ICMP 
 *						   client module.  Calling ICMPSendPing() 
 *						   will corrupt the other application's ping 
 *						   result.
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)
Public Function ICMPBeginUsage() As Boolean
   If ICMPFlags.bICMPInUse Then
      result = false
      Exit
   EndIf

   ICMPFlags.bICMPInUse = true
   result = true
End Function
(*
 *********************************************************************
 * Function:        void ICMPEndUsage(void)
 * PreCondition:    ICMPBeginUsage() was called by you and it 
 *					returned TRUE.
 * Input:           None
 * Output:          Your ownership of the ICMP module is released.  
 *					You can no longer use ICMPSendPing().
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)
Public Sub ICMPEndUsage()
   ICMPFlags.bICMPInUse = false
End Sub

// initialise...
#if defined(STACK_USE_ICMP) 
ICMPState = SM_IDLE
ICMPFlags.Val = 0
ICMPTimer = 0
Clear(ICMPHeader)
#endif

