{
*****************************************************************************
*  Name    : UDP.BAS                                                        *
*  Date    : 30/08/2007                                                     *
*  Version : 1.0 - Based around Microchip TCP/IP stack version 4.13         *
*  Notes   :                                                                *
* ========================================================================= *
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights reserved.     *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and distribute: *
* (i)  the Software when embedded on a Microchip microcontroller or digital *
*      signal controller product (Device) which is integrated into        *
*      Licensees product; or                                               *
* (ii) ONLY the Software driver source files ENC28J60.c and ENC28J60.h      *
*      ported to a non-Microchip device used in conjunction with a          *
*      Microchip ethernet controller for the sole purpose of interfacing    *
*      with the ethernet controller.                                        *
*                                                                           *
* You should refer to the license agreement accompanying this Software for  *
* additional information regarding your rights and obligations.             *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF   *
* ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY    *
* WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND  *
* NON-INFRINGEMENT. IN NO EVENT SHALL MICROCHIP BE LIABLE FOR ANY           *
* INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR   *
* LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR         *
* SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY   *
* DEFENSE THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER      *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT (INCLUDING *
* NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                            *
*****************************************************************************
}
Module UDP

Include "NETConfig.bas"
Include "NETTypes.bas"
Include "NETUtils.bas"
Include "IP.bas"
Include "MAC.bas"

Public Type 
   UDP_PORT = Word,
   UDP_SOCKET = Byte
   
Structure UDP_SOCKET_INFO
   remoteNode As NODE_INFO
   remotePort As UDP_PORT
   localPort As UDP_PORT
End Structure

Public Const
   INVALID_UDP_SOCKET      = ($ff),
   INVALID_UDP_PORT        = (0)

Public Dim
   activeUDPSocket As UDP_SOCKET,
   UDPSocketInfo(MAX_UDP_SOCKETS) As UDP_SOCKET_INFO,
   UDPTxCount As Word,
   UDPRxCount As Word

Structure UDP_HEADER
   SourcePort As UDP_PORT
   DestinationPort As UDP_PORT
   Length As Word
   Checksum As Word
End Structure

(*
 *********************************************************************
 * Macro:           UDPSetRxBuffer(a)
 * PreCondition:    None
 * Input:           a       - Offset
 * Output:          Next Read/Write access to receive buffer is
 *                  set to offset 'b'
 * Side Effects:    None
 * Note:            None
 ********************************************************************/
*)
Sub UDPSetRxBuffer(a As Word) 
   IPSetRxBuffer(a + SizeOf(UDP_HEADER))
End Sub
(*   
 *********************************************************************
 *	User Datagram Protocol (UDP) Communications Layer
 *  Module For Microchip TCP/IP Stack
 *	 -Provides unreliable, minimum latency transport of application 
 *    datagram (packet) oriented data
 *	 -Reference: RFC 768
 *
 *********************************************************************
*)

Const
   LOCAL_UDP_PORT_START_NUMBER  = (4096),
   LOCAL_UDP_PORT_END_NUMBER    = (8192)

Structure TFlags
   bFirstRead As Boolean
   bWasDiscarded As Boolean
End Structure
   
Dim
   LastPutSocket As UDP_SOCKET,
   wPutOffset As Word,
   Flags As TFlags,
   SocketWithRxData As UDP_SOCKET
   
(*
 *********************************************************************
 * Function:        void UDPSetTxBuffer(WORD wOffset)
 * PreCondition:    None
 * Input:           wOffset - Offset from begining of UDP packet data 
 *							  payload to set write pointer for 
 *							  UDPPut(), UDPPutArray(), etc.
 * Output:          None
 * Side Effects:    None
 * Overview:        Updates memory access pointers
 * Note:            None
 ********************************************************************/
*)
Sub UDPSetTxBuffer(wOffset As Word)
   IPSetTxBuffer(wOffset + SizeOf(UDP_HEADER))
   wPutOffset = wOffset
End Sub
(*
 *********************************************************************
 * Function:        void UDPClose(UDP_SOCKET S)
 * PreCondition:    UDPOpen() is already called
 * Input:           S       - Socket that is To be closed.
 * Output:          None
 * Side Effects:    None
 * Overview:        Given socket is marked As available For future
 *                  new communcations.
 * Note:            This Function does Not affect previous
 *                  active UDP socket designation.
 ********************************************************************/
*)
Public Sub UDPClose(S As UDP_SOCKET)
   If s = INVALID_UDP_SOCKET Then
      Exit
   EndIf
   UDPSocketInfo(S).localPort = INVALID_UDP_PORT
   UDPSocketInfo(S).remoteNode.IPAddr.Val = $00000000
End Sub
(*
 *********************************************************************
 * Function:        void UDPInit(void)
 * PreCondition:    None
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        Initializes internal variables.
 * Note:
 ********************************************************************/
*)
Public Sub UDPInit()
   Dim S As UDP_SOCKET

   S = 0
   While S < MAX_UDP_SOCKETS
      UDPClose(S)
      Inc(S)	
   Wend
   Flags.bWasDiscarded = true
End Sub
(*
 *********************************************************************
 * Function:        UDP_SOCKET UDPOpen(UDP_PORT localPort,
 *                                     NODE_INFO *remoteNode,
 *                                     UDP_PORT remotePort)
 * PreCondition:    UDPInit() is already called
 * Input:           remoteNode - Remote Node info such As MAC And IP
 *                               address
 *                               If NULL, broadcast node address is set.
 *                  remotePort - Remote Port To which To talk To
 *                               If INVALID_UDP_SOCKET, localPort is
 *                               opened For Listen.
 *                  localPort  - A valid Port number Or 0x0000 If 
 *								 dynamic assignment is desired.
 * Output:          On success: A UDP socket handle that can be used 
 *					For subsequent UDP API calls.
 *					On failure: INVALID_UDP_SOCKET
 * Side Effects:    None
 * Overview:        A UDP socket is allocated And parameters are stored.
 * Note:            If the localPort is Not specified, a unique Port 
 *					number is automatically assigned.
 ********************************************************************/
*)
// local statics...
Dim NextPort As Word
Public Function UDPOpen(localPort As UDP_PORT, ByRef remoteNode As NODE_INFO, remotePort As UDP_PORT) As UDP_SOCKET
   Dim S As UDP_SOCKET
   Dim P As UDPSocketInfo  // alias

   S = 0
   While S < MAX_UDP_SOCKETS
      If P(S).localPort = INVALID_UDP_PORT Then
         P(S).localPort = localPort	
         If localPort = $0000 Then
		    If NextPort > LOCAL_UDP_PORT_END_NUMBER Or NextPort < LOCAL_UDP_PORT_START_NUMBER Then
		       NextPort = LOCAL_UDP_PORT_START_NUMBER
		    EndIf
            P(S).localPort = NextPort
            Inc(NextPort)
         EndIf

         // If remoteNode is supplied, remember it.
         If @remoteNode <> 0 Then
            P(S).remoteNode = remoteNode
         Else
		   // else Set broadcast address
		   NETUtils.memset(@P(S).remoteNode, $FF, SizeOf(NODE_INFO))
		 EndIf

         P(S).remotePort   = remotePort

         // Mark this socket as active.
         // Once an active socket is set, subsequent operation can be
         // done without explicitely supply socket identifier.
         activeUDPSocket = S
         result = S
         Exit
      EndIf
      Inc(S)
   Wend
   result = INVALID_UDP_SOCKET
End Function
(*
 *********************************************************************
 * Function:        BOOL UDPIsPutReady(UDP_SOCKET S)
 * PreCondition:
 * Input:           S       - Socket that is To be loaded And made
 *                            an active UDP socket.
 * Output:          Word: Number of Bytes that can be put, 0 If none
 * Side Effects:    None
 * Overview:        None
 * Note:            This call sets given socket As an active UDP socket.
 ********************************************************************/
*)
Public Function UDPIsPutReady(S As UDP_SOCKET) As Word
   If LastPutSocket <> S Then
      LastPutSocket = S
      MACSetWritePtr(BASE_TX_ADDR + SizeOf(ETHER_HEADER) + SizeOf(IP_HEADER) + SizeOf(UDP_HEADER))
      UDPTxCount = 0
      wPutOffset = 0
   EndIf
   activeUDPSocket = S
   If Not MACIsTxReady() Then
      result = 0
      Exit
   EndIf
   result = MAC_TX_BUFFER_SIZE - SizeOf(IP_HEADER) - SizeOf(UDP_HEADER) - UDPTxCount
End Function
(*
 *********************************************************************
 * Function:        BOOL UDPPut(Byte v)
 * PreCondition:    UDPIsPutReady() > 0 With desired UDP socket
 *                  that is To be loaded.
 * Input:           v       - Data Byte To loaded into transmit buffer
 * Output:          TRUE If put successful
 *                  FALSE If transmit buffer is full And put failed
 * Side Effects:    None
 * Overview:        Given data Byte is put into UDP transmit buffer
 *                  And active UDP socket buffer Length is incremented
 *                  by one.
 *                  If buffer has become full, FALSE is returned.
 *                  Or Else TRUE is returned.
 * Note:            This Function loads data into an active UDP socket
 *                  As determined by previous call To UDPIsPutReady()
 ********************************************************************/
*)
Public Function UDPPut(v As Byte) As Boolean
	
   // See if we are out of transmit space.
   If wPutOffset >= (MAC_TX_BUFFER_SIZE - SizeOf(IP_HEADER) - SizeOf(UDP_HEADER)) Then
      result = false
      Exit
   EndIf

   // Load application data byte
   MACPut(v)
   Inc(wPutOffset)
   If wPutOffset > UDPTxCount Then
      UDPTxCount = wPutOffset
   EndIf
   result = true
End Function
(*
 *********************************************************************
 * Function:        Word UDPPutArray(Byte *cData, Word wDataLen)
 * PreCondition:    UDPIsPutReady() > 0 With desired UDP socket
 *                  that is To be loaded.
 * Input:           *cData - Pointer To data To write
 *					wDataLen - Length of data @ *cData To write
 * Output:          Word: Number of Bytes successfully placed in the 
 *						  UDP transmit buffer
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)
Public Function UDPPutArray(cData As Word, wDataLen As Word) As Word
   Dim wTemp As Word

   wTemp = (MAC_TX_BUFFER_SIZE - SizeOf(IP_HEADER) - SizeOf(UDP_HEADER)) - wPutOffset
   If wTemp < wDataLen Then
      wDataLen = wTemp
   EndIf   

   wPutOffset = wPutOffset + wDataLen
   If wPutOffset > UDPTxCount Then
      UDPTxCount = wPutOffset
   EndIf   

   // Load application data bytes
   MACPutArray(cData, wDataLen)
   result = wDataLen
End Function
(*
 *********************************************************************
 * Function:        Word UDPPutROMArray(ROM Byte *cData, Word wDataLen)
 * PreCondition:    UDPIsPutReady() > 0 With desired UDP socket
 *                  that is To be loaded.
 * Input:           *cData - Pointer To data To write
 *					wDataLen - Length of data @ *cData To write
 * Output:          Word: Number of Bytes successfully placed in the 
 *						  UDP transmit buffer
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)
Public Function UDPPutROMArray(cData As TROMAddress, wDataLen As Word) As Word
   Dim wTemp As Word

   wTemp = (MAC_TX_BUFFER_SIZE - SizeOf(IP_HEADER) - SizeOf(UDP_HEADER)) - wPutOffset
   If wTemp < wDataLen Then
      wDataLen = wTemp
   EndIf   

   wPutOffset = wPutOffset + wDataLen
   If wPutOffset > UDPTxCount Then
      UDPTxCount = wPutOffset
   EndIf   

   // Load application data bytes
   MACPutROMArray(cData, wDataLen)
   result = wDataLen
End Function
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Public Function UDPPutString(strData As String) As Word
   result = @strData + UDPPutArray(@strData, strlen(@strData))
End Function
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Public Function UDPPutString(strData As Word) As Word
   result = strData + UDPPutArray(strData, strlen(strData))
End Function
{
****************************************************************************
* Name    :                                                                *
* Purpose :                                                                *
****************************************************************************
}
Public Function UDPPutROMString(strData As TROMAddress) As TROMAddress
   result = strData + UDPPutROMArray(strData, strlenpgm(strData))
End Function
(*
 *********************************************************************
 * Function:        BOOL UDPFlush(void)
 * PreCondition:    UDPPut() is already called And desired UDP socket
 *                  is set As an active socket by calling
 *                  UDPIsPutReady().
 * Input:           None
 * Output:          All And any data associated With active UDP socket
 *                  buffer is marked As ready For transmission.
 * Side Effects:    None
 * Overview:        None
 * Note:            This Function transmit all data from
 *                  an active UDP socket.
 ********************************************************************/
*)
Public Sub UDPFlush()
   Dim h As UDP_HEADER

   // Wait for TX hardware to become available (finish transmitting 
   // any previous packet)
   While Not IPIsTxReady()
   Wend

   h.SourcePort        = NETUtils.swaps(UDPSocketInfo(activeUDPSocket).localPort)
   h.DestinationPort   = NETUtils.swaps(UDPSocketInfo(activeUDPSocket).remotePort)
   h.Length            = UDPTxCount + SizeOf(UDP_HEADER)
   
   // Do not swap h.Length yet.  It is needed in IPPutHeader.
   h.Checksum          = $0000

   MACSetWritePtr(BASE_TX_ADDR + SizeOf(ETHER_HEADER))

   // Load IP header.
   IPPutHeader(UDPSocketInfo(activeUDPSocket).remoteNode, IP_PROT_UDP, h.Length )

   // Now swap h.Length.
   h.Length            = swaps(h.Length)

   // Now load UDP header.
   MACPutArray(@h, SizeOf(h))

   // Update checksum.
   // TO BE IMPLEMENTED

   MACFlush()

   UDPTxCount = 0
   LastPutSocket = INVALID_UDP_SOCKET
End Sub
(*
 *********************************************************************
 * Function:        Word UDPIsGetReady(UDP_SOCKET S)
 * PreCondition:    UDPInit() is already called.
 * Input:           A valid UDP socket that is already "Listen"ed on
 *                  Or opened.
 * Output:          Word: Number of available Bytes that can be 
 *					retrieved.  0 If no data received on this socket.
 * Side Effects:    Given socket is set As an active UDP Socket.
 * Overview:        None
 * Note:            This Function automatically sets supplied socket
 *                  As an active socket.  Caller need Not call
 *                  explicit Function UDPSetActiveSocket().  All
 *                  subsequent calls will us this socket As an
 *                  active socket.
 ********************************************************************/
*)
Public Function UDPIsGetReady(S As UDP_SOCKET) As Word
   activeUDPSocket = s
   
   If SocketWithRxData <> s Then
      result = 0
      Exit
   EndIf
   result = UDPRxCount
End Function
(*
 *********************************************************************
 * Function:        void UDPDiscard(void)
 * PreCondition:    UDPInit() is already called
 * Input:           None
 * Output:          None
 * Side Effects:    None
 * Overview:        None
 * Note:            This Function discards an active UDP socket content.
 ********************************************************************/
*)
Public Sub UDPDiscard()
   If Not Flags.bWasDiscarded Then
      MACDiscardRx()
	  UDPRxCount = 0
	  Flags.bWasDiscarded = true
   EndIf
End Sub
(*
 *********************************************************************
 * Function:        BOOL UDPGet(Byte *v)
 * PreCondition:    UDPInit() is already called     And
 *                  UDPIsGetReady(S) > 0
 * Input:           v       - Buffer To receive UDP data Byte
 * Output:          TRUE    If a data Byte was read
 *                  FALSE   If no data Byte was read Or available
 * Side Effects:    None
 * Overview:        None
 * Note:            This Function fetches data from an active UDP
 *                  socket As set by UDPIsGetReady() call.
 ********************************************************************/
*)
Public Function UDPGet(ByRef v As Byte) As Boolean
   
   // Make sure that there is data to return
   If UDPRxCount = 0 Or SocketWithRxData <> activeUDPSocket Then
      result = false
      Exit
   EndIf

   // If if this very first read to packet, set MAC Rx Pointer to
   // begining of UDP data area.
   If Flags.bFirstRead Then
      Flags.bFirstRead = false
      UDPSetRxBuffer(0)
   EndIf

   v = MACGet()
   Dec(UDPRxCount)

   If UDPRxCount = 0 Then
      UDPDiscard()
   EndIf
   result = true
End Function
(*
 *********************************************************************
 * Function:        Word UDPGetArray(Byte *cData, Word wDataLen)
 * PreCondition:    UDPInit() is already called     And
 *                  UDPIsGetReady(S) > 0
 * Input:           *cData - Pointer To location To write retrieved Bytes
 *					wDataLen - Number of Bytes To retreive
 * Output:          Word - Number of Bytes written To cData
 * Side Effects:    None
 * Overview:        None
 * Note:            This Function fetches data from an active UDP
 *                  socket As set by UDPIsGetReady()
 ********************************************************************/
*)
Public Function UDPGetArray(cData As Word, wDataLen As Word) As Word

   // Make sure that there is data to return
   If UDPRxCount = 0 Or SocketWithRxData <> activeUDPSocket Then
      result = 0
      Exit
   EndIf

   // If this is the very first read of the packet, set MAC Rx Pointer to
   // beginig of UDP data area.
   If Flags.bFirstRead Then
      Flags.bFirstRead = false
      UDPSetRxBuffer(0)
   EndIf

   // Make sure we don't try to read more data than exists
   If UDPRxCount < wDataLen Then
      wDataLen = UDPRxCount
   EndIf   

   wDataLen = MACGetArray(cData, wDataLen)
   UDPRxCount = UDPRxCount - wDataLen

   If UDPRxCount = 0 Then
      UDPDiscard()
   EndIf
   result = wDataLen
End Function

(*
 *********************************************************************
 * Function:        UDP_SOCKET FindMatchingSocket(UDP_HEADER *h,
 *                                NODE_INFO *remoteNode,
 *                                IP_ADDR *localIP)
 * PreCondition:    UDP Segment header has been retrieved from buffer
 *                  The IP header has also been retrieved
 * Input:           remoteNode      - Remote node info from IP header
 *                  h               - header of UDP semgent.
 * Output:          matching UDP socket Or INVALID_UDP_SOCKET
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)
Const BROADCAST_ADDRESS = ($ffffffff)
Function FindMatchingSocket(ByRef h As UDP_HEADER, ByRef remoteNode As NODE_INFO, ByRef localIP As IP_ADDR) As UDP_SOCKET
   Dim S As UDP_SOCKET
   Dim partialMatch As UDP_SOCKET
   Dim P As UDPSocketInfo // alias

   partialMatch = INVALID_UDP_SOCKET

   S = 0
   While S < MAX_UDP_SOCKETS
      // This packet is said to be matching with current socket:
      // 1. If its destination port matches with our local port and
      // 2. Packet source IP address matches with socket remote IP address.
      //    OR this socket had transmitted packet with destination address as broadcast 
      //    (subnet or limited broadcast).
      
      If P(S).localPort = h.DestinationPort Then
         If (P(S).remotePort = h.SourcePort) Then
            If p(s).remoteNode.IPAddr.Val = remoteNode.IPAddr.Val Or 
               localIP.Val = $FFFFFFFF Or 
               localIP.Val = (AppConfig.MyIPAddr.Val Or (Not AppConfig.MyMask.Val)) Then
                  result = S
                  Exit
            EndIf
         EndIf
         partialMatch = S
      EndIf
      Inc(S)
   Wend

   If partialMatch <> INVALID_UDP_SOCKET Then
      P(partialMatch).remoteNode = remoteNode
      P(partialMatch).remotePort = h.SourcePort
   EndIf
   result = partialMatch
End Function

// Pseudo header as defined by rfc 793 (UDPProcess)
Structure PSEUDO_HEADER
   SourceAddress As IP_ADDR
   DestAddress As IP_ADDR
   Zero As Byte
   Protocol As Byte
   Length As Word
End Structure

// swap pseudo header (UDPProcess)
Sub SwapPseudoHeader(ByRef h As PSEUDO_HEADER)
   h.Length = NETUtils.swaps(h.Length)
End Sub   
(*
 *********************************************************************
 * Function:        BOOL UDPProcess(NODE_INFO* remoteNode,
 *                                  IP_ADDR *localIP,
 *                                  WORD len)
 * PreCondition:    UDPInit() is already called     AND
 *                  UDP segment is ready in MAC buffer
 * Input:           remoteNode      - Remote node info
 *					localIP			- Destination IP address of the packet
 *                  len             - Total length of UDP semgent.
 * Output:          TRUE if a valid packet is waiting
 *                  FALSE if the packet was discarded
 * Side Effects:    None
 * Overview:        None
 * Note:            None
 ********************************************************************/
*)
Public Function UDPProcess(ByRef remoteNode As NODE_INFO, ByRef localIP As IP_ADDR, len As Word) As Boolean
   Dim h As UDP_HEADER
   Dim S As UDP_SOCKET
   Dim pseudoHeader As PSEUDO_HEADER
   Dim checksums As DWORD_VAL
   
   SocketWithRxData = INVALID_UDP_SOCKET
   UDPRxCount = 0
   Flags.bFirstRead = false
	
   // Retrieve UDP header.
   MACGetArray(@h, SizeOf(h))

   h.SourcePort        = NETUtils.swaps(h.SourcePort)
   h.DestinationPort   = NETUtils.swaps(h.DestinationPort)
   h.Length            = NETUtils.swaps(h.Length) - SizeOf(UDP_HEADER)

   // See if we need to validate the checksum field (0x0000 is disabled)
   If h.Checksum <> 0 Then
      // Calculate IP pseudoheader checksum.
	  pseudoHeader.SourceAddress	= remoteNode.IPAddr
	  pseudoHeader.DestAddress.Val	= localIP.Val
	  pseudoHeader.Zero				= $0
	  pseudoHeader.Protocol			= IP_PROT_UDP
	  pseudoHeader.Length			= len

      SwapPseudoHeader(pseudoHeader)
	
	  checksums.w(0) =  Not CalcIPChecksum(@pseudoHeader, SizeOf(pseudoHeader))
		
      // Now calculate UDP packet checksum in NIC RAM -- should match pseudoHeader
	  IPSetRxBuffer(0)
      checksums.w(1) = MAC.CalcIPBufferChecksum(len)
	
      If checksums.w(0) <> checksums.w(1) Then
         MACDiscardRx()
         result = false
         Exit
      EndIf
   EndIf
   
   S = FindMatchingSocket(h, remoteNode, localIP)
   If S = INVALID_UDP_SOCKET Then
      // If there is no matching socket, There is no one to handle
      // this data.  Discard it.
      MACDiscardRx()
      result = false
      Exit
   Else
      SocketWithRxData = s
      UDPRxCount = h.Length
      Flags.bFirstRead = true
	  Flags.bWasDiscarded = false
   EndIf
   result = true
End Function

// initialise...
#if defined(STACK_USE_UDP)
LastPutSocket = INVALID_UDP_SOCKET
NextPort = 0
SocketWithRxData = INVALID_UDP_SOCKET
UDPRxCount = 0
UDPTxCount = 0
#endif




