//
//-------------------------------------------------------------------------
//  Name    : ENC.bas
//  Author  : Jerry Messina
//  Date    : 9/23/2015
//  Version : 1.0.0
//
//  ENC interface module (MSSP master SPI)
//  this module defaults to using MSSP1 with an SPI_CLK of SPI_OSC_DIV4
//
//  ver 1.0.0
//-------------------------------------------------------------------------
//
module ENC

// IO pin for ENC_CS spi chip select output (required)
#if IsOption(ENC_CS) And Not IsValidPortPin(ENC_CS) 
   #error ENC_CS, "Invalid option. CS must be a valid port pin."
#endif
'#option ENC_CS = PORTD.6

// IO pin for ENC_RST reset (optional, output)
#if IsOption(ENC_RST) And Not IsValidPortPin(ENC_RST) 
   #error ENC_RST, "Invalid option. RST must be a valid port pin."
#endif
'#option ENC_RST = PORTD.5

// IO pin for ENC_INT interrupt pin (optional, polled input)
#if IsOption(ENC_INT) And Not IsValidPortPin(ENC_INT) 
   #error ENC_INT, "Invalid option. INT must be a valid port pin."
#endif
'#option ENC_INT = PORTD.5

// set default MSSP selection
#option ENC_MSSP = 1

// check device for MSSP support
#if (_mssp = 0)
   #error _device + " does not support MSSP"
#elseif (_mssp = 1) and (ENC_MSSP > 1)
   #error _device + " does not support MSSP2"
#endif

#if (ENC_MSSP = 1)
// MSSP1 control registers
dim
    MSSPCON1      as SSPCON1,
    MSSPSTAT      as SSPSTAT,
    MSSPBUF       as SSPBUF

// MSSP1 pins
dim
    SCK     as PORTC.3,     // SCK1 clock output pin
    SDI     as PORTC.4,     // SDI1 data in (MISO)
    SDO     as PORTC.5      // SDO1 data out (MOSI)

// MSSP1 interrupt flag bit (polled only)
dim 
    SSPIF as PIR1.bits(3),      // SSP1IF
    SSPIE as PIE1.bits(3)       // SSP1IE

#elseif (ENC_MSSP = 2)
// MSSP2 control registers
dim
    MSSPCON1      as SSP2CON1,
    MSSPSTAT      as SSP2STAT,
    MSSPBUF       as SSP2BUF

// MSSP2 pins (varies with device)
#if (_device = 18F26K22)
dim
    SCK     as PORTB.1,     // SCK2 clock output pin
    SDI     as PORTB.2,     // SDI2 data in (MISO)
    SDO     as PORTB.3      // SDO2 data out (MOSI)
#elseif (_device = 18F46K22)
dim
    SCK     as PORTD.0,     // SCK2 clock output pin
    SDI     as PORTD.1,     // SDI2 data in (MISO)
    SDO     as PORTD.4      // SDO2 data out (MOSI)
#elseif (_device = 18F8722) or (_device = 18F87K22)
dim
    SCK     as PORTD.6,     // SCK2 clock output pin
    SDI     as PORTD.5,     // SDI2 data in (MISO)
    SDO     as PORTD.4      // SDO2 data out (MOSI)
#else
  #error "add MSSP2 SPI pin definitions for the " + _device
#endif

// MSSP2 interrupt flag bits
dim 
    SSPIF as PIR3.bits(7),      // SSP2IF
    SSPIE as PIR3.bits(7)       // SSP2IE

#else   // not MSSP1 or MSSP2
  #error "invalid MSSP selection"
#endif

// SSPSTAT bits
const
    SMP = 7,            // read sample mode
    CKE = 6,            // clock edge control
    BF  = 0             // buffer full (receive and transmit)    

// SSPCON1 bits, master mode only...
const
    WCOL  = 7,        // write collision Detect
    SSPOV = 6,        // receive overflow
    SSPEN = 5,        // synchronous port enable
    CKP   = 4,        // clock polarity
    SSPM3 = 3,        // synchronous mode select bits
    SSPM2 = 2,
    SSPM1 = 1,
    SSPM0 = 0 

// standard SPI bus modes
public const
    SPI_MODE_0 = 0,        // CKP=0, CKE=1... clock idle low, mosi changes on falling clk
    SPI_MODE_1 = 1,        // CKP=0, CKE=0... clock idle low, mosi changes on rising clk
    SPI_MODE_2 = 2,        // CKP=1, CKE=1... clock idle high, mosi changes on rising clk
    SPI_MODE_3 = 3         // CKP=1, CKE=0... clock idle high, mosi changes on falling clock

public const
    SPI_MODE_00 = SPI_MODE_0,
    SPI_MODE_01 = SPI_MODE_1,
    SPI_MODE_10 = SPI_MODE_2,
    SPI_MODE_11 = SPI_MODE_3

public const
    // SSPM3:SSPM0 SPI clock mode select bits
    SPI_OSC_DIV4         = %0000,   // master mode clk=FOSC/4
    SPI_OSC_DIV16        = %0001,   // master mode clk=FOSC/16
    SPI_OSC_DIV64        = %0010,   // master mode clk=FOSC/64
    SPI_OSC_TMR2         = %0011,   // master mode clk=TMR2/2

    // clock idle and edge settings  (SPI Mode compatible)
    SPI_RISING_EDGE      = $40, // data transmitted on rising edge of clock
    SPI_FALLING_EDGE     = $00, // data transmitted on falling edge of clock
    SPI_IDLE_HIGH        = $10, // idle state for clock is high
    SPI_IDLE_LOW         = $00, // idle state for clock is low

    // RX data sampling settings  (SSPStatus.6 - SMP)
    SPI_SMP_END          = $80, // input data sampled at end of data output time
    SPI_SMP_MID          = $00  // input data sampled at middle of data output time

// set module default spi clock rate
//   enc28J60 SCK max = 20MHz
//   encx24J600 SCK max = 14MHz
// some devices have addtl clock modes... be sure to check the datasheet carefully
#if (_device = 18F87K22)
public const
    SPI_OSC_SSPADD       = %1000,   // master mode clk=FOSC/(4*(SSPADD+1))
    SPI_OSC_DIV8         = %1010    // master mode clk=FOSC/8
#elseif (_device = 18F26K22) or (_device = 18F46K22)
public const
    SPI_OSC_SSPADD       = %1000    // master mode clk=FOSC/(4*(SSPADD+1))
#endif

#if (_device = 18F87K22) and (_clock = 64)
const
    SPI_DEFAULT_CLK = SPI_OSC_DIV8
#else
const
    SPI_DEFAULT_CLK = SPI_OSC_DIV4
#endif

//
//------------------------------------------------------------------------------
// ENC IO pins (CS, and optional INT, RESET)
//------------------------------------------------------------------------------
//
// ENC IO pins
dim ENC_CS_IO as ENC_CS.ENC_CS@

#if isoption(ENC_RST)    
dim ENC_RST_IO as ENC_RST.ENC_RST@
#endif

#if isoption(ENC_INT)
dim ENC_INT_IO as ENC_INT.ENC_INT@
#endif

// init IO pins
public inline sub InitIO()
    high(ENC_CS_IO)         // init CS output high
  #if isoption(ENC_RST)
    low(ENC_RST_IO)         // assert RESET output low
    delayus(1)              // Tresetlow = 400ns min
    ENC_RST_IO = 1          // release reset
  #endif
  #if isoption(ENC_INT)
    input(ENC_INT_IO)       // INT polled input
  #endif  
end sub

inline sub nop()
    asm
        NOP
    end asm
end sub

// ENC CS control
// the enc28J60 and x24J600 have a Tcss chip select to sck setup time of 50ns
public inline sub AssertCS()
    ENC_CS_IO = 0
    'nop()          // TCSS setup time (optional)
end sub

// the enc28J60 has a Tcsh chip select hold time of 210ns when accessing the
// MAC and MII registers, and a Tcsd chip select disable time of 50ns
// the encx24J600 has a Tcsh hold time of 50ns, and a Tcsd of 20ns
//   10M = 400ns Tcy
//   20M = 200ns Tcy
//   40M = 100ns Tcy
//   50M =  80ns Tcy
//   64M =  64ns Tcy
public inline sub DeassertCS()
    'nop()           // TCSH hold time
  #if isoption(ETH_MAC) and (ETH_MAC = ETHXXJ60)
  #if (_clock > 20)
    nop()
  #endif
  #if (_clock > 40)
    nop()
    nop()
  #endif
  #endif    
    ENC_CS_IO = 1
    nop()           // TCSD disable time
end sub

// ENC RST control
public inline sub AssertRST()
  #if isoption(ENC_RST)    
    ENC_RST_IO = 0
  #endif  
end sub

public inline sub DeassertRST()
  #if isoption(ENC_RST)    
    ENC_RST_IO = 1
  #endif  
end sub

// ENC INT handing
// the enc INT pin can be used as a signal to the uC that the enc
// needs to be polled (saves on SPI traffic)
public inline function GetINT() as boolean
    // assume int pin is asserted
    // this way, if it's not actually used then the ENC will be polled 
    // for status like it normally would
    result = true
  #if isoption(ENC_INT)
    if (ENC_INT_IO = 1) then    // INT pin is NOT asserted
        result = false
    endif        
  #endif
end function

//
//------------------------------------------------------------------------------
// SetAsMaster()
// Set SPI to master mode
//      : SPI_OSC_DIV4, SPI_OSC_DIV8(DEFAULT), SPI_OSC_DIV16, SPI_OSC_DIV64, spiOscTmr2
//------------------------------------------------------------------------------
//
public inline sub set_spi_master(pMode as byte = SPI_DEFAULT_CLK)
    MSSPCON1.bits(SSPEN) = 0  // disable port for now

    output(SCK)
    output(SDO)
    input(SDI)

    MSSPCON1 = MSSPCON1 and $C0       // clear WCOL and SSPOV
    MSSPCON1 = MSSPCON1 or pMode
end sub

//
//------------------------------------------------------------------------------
// SetSample()
// Sets when the SPI input data in sampled (SMP)
//      : SPI_SMP_END, SPI_SMP_MID.
//      : For slave mode, sample should always be SPI_SMP_MID
//------------------------------------------------------------------------------
//
inline sub SetSample(pSample as byte)
    if (pSample = 0) then
        MSSPSTAT.bits(SMP) = 0             // SPI_SMP_MID
    else
        MSSPSTAT.bits(SMP) = 1             // SPI_SMP_END
    endif
end sub

//
//------------------------------------------------------------------------------
// SetClock()
// Set SPI idle state and data transmission clock edge
//      : pIdle -> SPI_IDLE_HIGH, SPI_IDLE_LOW (CKP)
//      : pEdge -> SPI_RISING_EDGE, SPI_FALLING_EDGE (CKE)
//------------------------------------------------------------------------------
//
inline sub SetClock(pIdle as byte, pEdge as byte)
    // clock polarity CKP
    MSSPCON1 = MSSPCON1 and $EF
    MSSPCON1 = MSSPCON1 or pIdle

    if (pIdle = SPI_IDLE_HIGH) then
        pEdge = not pEdge and $BF
    endif

    MSSPSTAT = MSSPSTAT and $BF
    MSSPSTAT = MSSPSTAT or pEdge
end sub

// standard SPI bus modes
macro set_mode(m)
    if (m = SPI_MODE_1) then
        MSSPCON1.bits(CKP) = 0
        MSSPSTAT.bits(CKE) = 0
    elseif (m = SPI_MODE_2) then
        MSSPCON1.bits(CKP) = 1
        MSSPSTAT.bits(CKE) = 1
    elseif (m = SPI_MODE_3) then
        MSSPCON1.bits(CKP) = 1
        MSSPSTAT.bits(CKE) = 0
    else    // default to SPI mode 0 (0,0)
        MSSPCON1.bits(CKP) = 0
        MSSPSTAT.bits(CKE) = 1
    endif

    // sample input data at middle of data out time
    setsample(SPI_SMP_MID)
end macro

//
//------------------------------------------------------------------------------
// transfer()
//  does a proper SPI transaction (write and read)
//------------------------------------------------------------------------------
//
// note: these routines pass and return the parameters in WREG, so you might
// have to be careful how they're used
public inline function transfer(b as WREG) as WREG
    MSSPBUF = b
    // wait for byte to be sent/received
    while (MSSPSTAT.bits(BF) = 0)
    end while
    // SSPBUF return value will get the byte that just came in (and clears BF)
    result = MSSPBUF
end function

// short-hand alias for a read transfer (automatically writes a dummy 0)
public inline function transfer() as WREG
    result = transfer(0)
end function

{
// conventional parameter passing
public inline function transfer(b as byte) as byte
    MSSPBUF = b
    // wait for byte to be sent/received
    while (MSSPSTAT.bits(BF) = 0)
    end while
    // SSPBUF return value will get the byte that just came in (and clears BF)
    result = MSSPBUF
end function

// short-hand alias for a read transfer (automatically writes a dummy 0)
public inline function transfer() as byte
    result = transfer(0)
end function
}

//
//------------------------------------------------------------------------------
// burst_read()
//------------------------------------------------------------------------------
//
public inline sub burst_read(dest as word, len as word)
    dim fsr as FSR0,
        indf as INDF0,
        postinc as POSTINC0

    fsr = dest              // setup dest addr
    if (fsr = 0) then       // check for valid dest addr...
        exit
    endif        

    // read the data
    while (len <> 0)
        postinc = transfer()        // send a dummy byte to receive a byte
        dec(len)
    end while
end sub

//
//------------------------------------------------------------------------------
// burst_write()
//------------------------------------------------------------------------------
//
public inline sub burst_write(src as word, len as word)
    dim fsr as FSR0,
        indf as INDF0,
        postinc as POSTINC0

    fsr = src                   // setup src addr
    // write the data
    while (len <> 0)
        transfer(postinc)     // send byte
        dec(len)
    end while
end sub

public inline sub burst_write_from_rom(src as TABLEPTR, len as word)
    while (len <> 0)
        asm                         // read the src data
            TBLRD*+
        end asm
        transfer(TABLAT)          // send byte
        dec(len)
    end while
end sub

//
//------------------------------------------------------------------------------
// initialize port
//------------------------------------------------------------------------------
//
dim bInitialized as MSSPCON1.booleans(SSPEN)
public inline function initialized() as bInitialized
end function

public sub initialize(clk_sel as byte=SPI_DEFAULT_CLK)
    // setup control signals (CS, RST, INT)
    InitIO()
    
    // set SPI configuration
    set_spi_master(clk_sel)
    set_mode(SPI_MODE_0)

    // enable port
    SSPIF = 0
    MSSPCON1.bits(SSPEN) = 1
end sub

//
//------------------------------------------------------------------------------
// module initialization
//------------------------------------------------------------------------------
//

// disable the MSSP
MSSPCON1.bits(SSPEN) = 0
// disable intr
SSPIE = 0

end module

