{
*****************************************************************************
*  Name    : ETH97J60.BAS                                                   *
*  Date    : 04/04/2015                                                     *
*  Version : 1.1 - Based around Microchip ETH97J60.c version 5.36           *
*  Notes   :                                                                *
* ========================================================================= *
*   Medium Access Control (MAC) Layer for Microchip PIC18F97J60 family       *
*  Module for Microchip TCP/IP Stack                                        *
*    -Provides access to PIC18F97J60 family Ethernet controller              *
*    -Reference: PIC18F97J60 Family data sheet, IEEE 802.3 Standard          *
*                                                                           *
*****************************************************************************
* Software License Agreement                                                *
*                                                                           *
* Copyright  2002-2007 Microchip Technology Inc.  All rights               *
* reserved.                                                                 *
*                                                                           *
* Microchip licenses to you the right to use, modify, copy, and             *
* distribute:                                                               *
* (i)  the Software when embedded on a Microchip microcontroller or         *
*      digital signal controller product (Device) which is                *
*      integrated into Licensees product  or                               *
* (ii) ONLY the Software driver source files ENC28J60.c and                 *
*      ENC28J60.h ported to a non-Microchip device used in                  *
*      conjunction with a Microchip ethernet controller for the             *
*      sole purpose of interfacing with the ethernet controller.            *
*                                                                           *
* You should refer to the license agreement accompanying this               *
* Software for additional information regarding your rights and             *
* obligations.                                                              *
*                                                                           *
* THE SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT               *
* WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT        *
* LIMITATION, ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A                *
* PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL         *
* MICROCHIP BE LIABLE FOR ANY INCIDENTAL, SPECIAL, INDIRECT OR              *
* CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF                 *
* PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS       *
* BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE                *
* THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER              *
* SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT            *
* (INCLUDING NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE.                 *
*****************************************************************************
}                                                                             
Module ETH97J60

Include "NETConfig.bas"
Include "NETTypes.bas"
Include "NETUtils.bas"
Include "Tick.bas"
Include "system.bas"

// *********************** D E F I N I T I O N S ******************************
// Since the Ethernet PHY doesn't support auto-negotiation, full-duplex mode is
// not compatible with most switches/routers.  If a dedicated network is used
// where the duplex of the remote node can be manually configured, you may
// change this configuration.  Otherwise, half duplex should always be used
// note: this was originally done as a '#define'
#option ETH_DUPLEX = HALF_DUPLEX
#if (ETH_DUPLEX = HALF_DUPLEX)
  // default setting
#elseif (ETH_DUPLEX = FULL_DUPLEX)
  #warning "ETH FULL_DUPLEX may cause problems"
#else
  #error "invalid ETH_DUPLEX setting. must be FULL_DUPLEX or HALF_DUPLEX"
#endif

// Rest are Receive Buffers
Public Const
    MAC_IP      = ($00),
    MAC_ARP     = ($06),
    MAC_UNKNOWN = ($FF)
   
Const
    ETHER_IP     = ($00),
    ETHER_ARP    = ($06)

// A generic structure representing the Ethernet header starting all Ethernet
// frames
Public Structure MAC_ETHER_HEADER
    DestMACAddr     As MAC_ADDR
    SourceMACAddr   As MAC_ADDR
    _Type           As WORD_VAL
End Structure
 
  // TXSTATUS bits union
Structure TXBits
    v(7) As Byte
    // bytes 0 & 1 (word)
    ByteCount           As v(0).AsWord      // 16
    // byte 2
    _CollisionCount     As v(2)             // bits 0-3
    CRCError            As v(2).4
    LengthCheckError    As v(2).5
    LengthOutOfRange    As v(2).6
    Done                As v(2).7
    // byte 3
    Multicast           As v(3).0
    Broadcast           As v(3).1
    PacketDefer         As v(3).2
    ExcessiveDefer      As v(3).3
    MaximumCollisions   As v(3).4
    LateCollision       As v(3).5
    Giant               As v(3).6
    Underrun            As v(3).7
    // bytes 4 & 5 (word)
    BytesTransmittedOnWire As v(4).AsWord   // 16
    // byte 6
    ControlFrame        As v(6).0
    PauseControlFrame   As v(6).1
    BackpressureApplied As v(6).2
    VLANTaggedFrame     As v(6).3
    Zeros               As v(6).4           // bits 4-7
End Structure

// TXSTATUS...
Structure TXStatus
    v(7) As Byte
    Bits As TXBits Union
End Structure

// RXSTATUS bits union
Structure RXBits
    v(4) As Byte
    // bytes 0 & 1 (word)
    ByteCount           As v(0).AsWord      // 16
    // byte 2
    PreviouslyIgnored   As v(2).0
    RXDCPreviouslySeen  As v(2).1
    CarrierPreviouslySeen As v(2).2
    CodeViolation       As v(2).3
    CRCError            As v(2).4
    LengthCheckError    As v(2).5
    LengthOutOfRange    As v(2).6
    ReceiveOk           As v(2).Booleans(7)
    // byte 3
    Multicast           As v(3).0
    Broadcast           As v(3).1
    DribbleNibble       As v(3).2
    ControlFrame        As v(3).3
    PauseControlFrame   As v(3).4
    UnsupportedOpcode   As v(3).5
    VLANType            As v(3).6
    zero                As v(3).7
End Structure   

// RXSTATUS
Structure RXSTATUS
   v(4)     As Byte
   Bits     As RXBits Union
End Structure

// A header appended at the start of all RX frames by the hardware
Structure ENC_PREAMBLE
    NextPacketPointer   As Word
    StatusVector        As RXSTATUS
    DestMACAddr         As MAC_ADDR
    SourceMACAddr       As MAC_ADDR
    _Type               As WORD_VAL
End Structure

// Internal MAC level variables and flags
Dim
    NextPacketLocation As WORD_VAL,
    CurrentPacketLocation As WORD_VAL,
    WasDiscarded As Boolean

// Time of last transmission; used for determining when TX hardware may need software intervention
// (97J60 errata #7 transmit deadlock)
Dim
    wTXWatchdog As Word

// ethernet buffer sizes and locations
Public Const
    MAC_TX_BUFFER_SIZE = _MAC_TX_BUFFER_SIZE,
    RAMSIZE = _RAMSIZE,
    TXSTART = _TXSTART,
    RXSTART = _RXSTART,
    RXSTOP = _RXSTOP,
    RXSIZE = _RXSIZE,
    BASE_TX_ADDR = _BASE_TX_ADDR,
    BASE_TCB_ADDR = _BASE_TCB_ADDR,
    BASE_HTTPB_ADDR = _BASE_HTTPB_ADDR

//
// PHY Register Locations
//
Public Const
   PHCON1   = $00,
   PHSTAT1  = $01,
   PHCON2   = $10,
   PHSTAT2  = $11,
   PHIE     = $12,
   PHIR     = $13,
   PHLCON   = $14

// PHCON1 bits
Structure TPHCON1
    Val     As Word
    PDPXMD  As Val.8
End Structure

// PHSTAT1 bits
Structure TPHSTAT1
    Val     As Word
    LLSTAT  As Val.booleans(2)
End Structure

// PHCON2 bits
Structure TPHCON2
    Val     As Word
    RXAPDIS As Val.4
    HLDDIS  As Val.8
    FRCLNK  As Val.14
End Structure

// PHSTAT2 bits
Structure TPHSTAT2
    Val     As Word
    LSTAT   As Val.10
    COLSTAT As Val.11
    RXSTAT  As Val.12
    TXSTAT  As Val.13
End Structure

// PHIE bits
Structure TPHIE
    Val     As Word
    PGEIE   As Val.1
    PLNKIE  As Val.4
End Structure

// PHIR bits
Structure TPHIR
    Val     As Word
    PGIF    As Val.2
    PLNKIF  As Val.4
End Structure

// PHLCON bits
Structure TPHLCON
    Val     As Word
    STRCH   As Val.1
    LFRQ0   As Val.2
    LFRQ1   As Val.3
    LBCFG0  As Val.4
    LBCFG1  As Val.5
    LBCFG2  As Val.6
    LBCFG3  As Val.7
    LACFG0  As Val.8
    LACFG1  As Val.9
    LACFG2  As Val.10
    LACFG3  As Val.11
End Structure

Public Structure PHYREG
    Val         As WORD_VAL
    PHCON1bits  As TPHCON1 Union
    PHSTAT1bits As TPHSTAT1 Union
    PHCON2bits  As TPHCON2 Union
    PHSTAT2bits As TPHSTAT2 Union
    PHIEbits    As TPHIE Union
    PHIRbits    As TPHIR Union
    PHLCONbits  As TPHLCON Union
End Structure

{
*******************************************************************************
* Individual Register Bits
********************************************************************************
}
// ETH/MAC/MII bits
// EIE bits ----------
Const
    PKTIE   = 6,
    DMAIE   = 5,
    LINKIE  = 4,
    TXIE    = 3,
    TXERIE  = 1,
    RXERIE  = 0,

    EIE_PKTIE   = (1<<PKTIE),
    EIE_DMAIE   = (1<<DMAIE),
    EIE_LINKIE  = (1<<LINKIE),
    EIE_TXIE    = (1<<TXIE),
    EIE_TXERIE  = (1<<TXERIE),
    EIE_RXERIE  = (1<<RXERIE)

// EIR bits ----------
Const
    PKTIF   = 6,
    DMAIF   = 5,
    LINKIF  = 4,
    TXIF    = 3,
    TXERIF  = 1,
    RXERIF  = 0,

    EIR_PKTIF   = (1<<PKTIF),
    EIR_DMAIF   = (1<<DMAIF),
    EIR_LINKIF  = (1<<LINKIF),
    EIR_TXIF    = (1<<TXIF),
    EIR_TXERIF  = (1<<TXERIF),
    EIR_RXERIF  = (1<<RXERIF)
   
// ESTAT bits ---------
Const
    BUFER  = 6,
    RXBUSY = 2,
    TXABRT = 1,
    PHYRDY = 0,

    ESTAT_BUFER  = (1<<BUFER),
    ESTAT_RXBUSY = (1<<RXBUSY),
    ESTAT_TXABRT = (1<<TXABRT),
    ESTAT_PHYRDY = (1<<PHYRDY)
   
// ECON2 bits --------
Const
    AUTOINC = 7,
    PKTDEC  = 6,
    ETHEN   = 5,

    ECON2_AUTOINC = (1<<AUTOINC),
    ECON2_PKTDEC  = (1<<PKTDEC),
    ECON2_ETHEN   = (1<<ETHEN)
   
// ECON1 bits --------
Const
    TXRST   = 7,
    RXRST   = 6,
    DMAST   = 5,
    CSUMEN  = 4,
    TXRTS   = 3,
    RXEN    = 2,

    ECON1_TXRST   = (1<<TXRST),
    ECON1_RXRST   = (1<<RXRST),
    ECON1_DMAST   = (1<<DMAST),
    ECON1_CSUMEN  = (1<<CSUMEN),
    ECON1_TXRTS   = (1<<TXRTS),
    ECON1_RXEN    = (1<<RXEN)
   
// ERXFCON bits ------
Const
    UCEN  = 7,
    ANDOR = 6,
    CRCEN = 5,
    PMEN  = 4,
    MPEN  = 3,
    HTEN  = 2,
    MCEN  = 1,
    BCEN  = 0,
    
    ERXFCON_UCEN  = (1<<UCEN),
    ERXFCON_ANDOR = (1<<ANDOR),
    ERXFCON_CRCEN = (1<<CRCEN),
    ERXFCON_PMEN  = (1<<PMEN),
    ERXFCON_MPEN  = (1<<MPEN),
    ERXFCON_HTEN  = (1<<HTEN),
    ERXFCON_MCEN  = (1<<MCEN),
    ERXFCON_BCEN  = (1<<BCEN)
   
// MAC AND MII REGISTERS
// MACON1 bits --------
Const
    TXPAUS  = 3,
    RXPAUS  = 2,
    PASSALL = 1,
    MARXEN  = 0,

    MACON1_TXPAUS  = (1<<TXPAUS),
    MACON1_RXPAUS  = (1<<RXPAUS),
    MACON1_PASSALL = (1<<PASSALL),
    MACON1_MARXEN  = (1<<MARXEN)
   
// MACON3 bits --------
Const
    PADCFG2 = 7,
    PADCFG1 = 6,
    PADCFG0 = 5,
    TXCRCEN = 4,
    PHDREN  = 3,
    HFRMEN  = 2,
    FRMLNEN = 1,
    FULDPX  = 0,
    
    MACON3_PADCFG2 = (1<<PADCFG2),
    MACON3_PADCFG1 = (1<<PADCFG1),
    MACON3_PADCFG0 = (1<<PADCFG0),
    MACON3_TXCRCEN = (1<<TXCRCEN),
    MACON3_PHDREN  = (1<<PHDREN),
    MACON3_HFRMEN  = (1<<HFRMEN),
    MACON3_FRMLNEN = (1<<FRMLNEN),
    MACON3_FULDPX  = (1<<FULDPX)
   
// MACON4 bits --------
Const
    DEFER   = 6,

    MACON4_DEFER   = (1<<DEFER)
   
// MICMD bits ---------
Const
    MIISCAN  = 1,
    MIIRD    = 0,

    MICMD_MIISCAN  = (1<<MIISCAN),
    MICMD_MIIRD    = (1<<MIIRD)

// MISTAT bits --------
Const
    NVALID  = 2,
    SCAN    = 1,
    BUSY    = 0,

    MISTAT_NVALID  = (1<<NVALID),
    MISTAT_SCAN    = (1<<SCAN),
    MISTAT_BUSY    = (1<<BUSY)
   
// EFLOCON bits -----
Const
    FCEN1  = 1,
    FCEN0  = 0,

    EFLOCON_FCEN1  = (1<<FCEN1),
    EFLOCON_FCEN0  = (1<<FCEN0)

// PHY bits
// PHCON1 bits ----------
Const
    PHCON1_PDPXMD   As Word = (1<<8),

// PHSTAT1 bits --------
    PHSTAT1_LLSTAT  As Word = (1<<2),

// PHCON2 bits ----------
    PHCON2_FRCLNK   As Word = (1<<14),
    PHCON2_HDLDIS   As Word = (1<<8),
    PHCON2_RXAPDIS  As Word = (1<<4),

// PHSTAT2 bits --------
    PHSTAT2_TXSTAT  As Word = (1<<13),
    PHSTAT2_RXSTAT  As Word = (1<<12),
    PHSTAT2_COLSTAT As Word = (1<<11),
    PHSTAT2_LSTAT   As Word = (1<<10),

// PHIE bits -----------
    PHIE_PLNKIE     As Word = (1<<4),
    PHIE_PGEIE      As Word = (1<<1),

// PHIR bits -----------
    PHIR_PLNKIF     As Word = (1<<4),
    PHIR_PGIF       As Word = (1<<2),

// PHLCON bits -------
    PHLCON_LACFG3   As Word = (1<<11),
    PHLCON_LACFG2   As Word = (1<<10),
    PHLCON_LACFG1   As Word = (1<<9),
    PHLCON_LACFG0   As Word = (1<<8),
    PHLCON_LBCFG3   As Word = (1<<7),
    PHLCON_LBCFG2   As Word = (1<<6),
    PHLCON_LBCFG1   As Word = (1<<5),
    PHLCON_LBCFG0   As Word = (1<<4),
    PHLCON_LFRQ1    As Word = (1<<3),
    PHLCON_LFRQ0    As Word = (1<<2),
    PHLCON_STRCH    As Word = (1<<1)

// MAC and MII REGISTER access
// note 1: Do not access the MAC and MII SFRs unless the Ethernet module is enabled
//      (ETHEN = 1)
// note 2: Back-to-back accesses of MAC or MII registers are not supported. Between any
//      instruction which addresses a MAC or MII register, at least one  NOP or other
//      instruction must be executed.

// EDATA register access
// note: due to a PIC18F97J60 bug, you must use the MOVFF instruction to
// write to EDATA or else the read pointer (ERDPT) will inadvertently increment
// this really isn't a bug (see 97J60 datasheet sect 19.2.1.1) 
public macro WRITE_EDATA(b)
    PRODL = b
    asm
        movff   PRODL, EDATA
    end asm
end macro

// high byte/low byte access (replaces lo/hi functions)
macro lowbyte(w)
    (w and $00ff)
end macro

macro highbyte(w)
    (w >> 8)
end macro

{
******************************************************************************
* Function:        void MACPut(BYTE val)
*
* PreCondition:    EWRPT must point to the location to begin writing.
* Input:           Byte to write into the Ethernet buffer memory
* Output:          None
* Side Effects:    None
* Overview:        Writes to the EDATA register, which will indirectly
*                   increment EWRPTH:EWRPTL.
* Note:            None
*****************************************************************************
}
Public Sub MACPut(val As WREG)
    WRITE_EDATA(val)
End Sub

{
******************************************************************************
* Function:        Byte MACGet()
* PreCondition:    ERDPT must point to the place to read from.
* Input:           None
* Output:          Byte read from the Ethernet's buffer RAM
* Side Effects:    None
* Overview:        MACGet returns the byte pointed to by ERDPT and
*                  increments ERDPT so MACGet() can be called again.  The
*                  increment will follow the receive buffer wrapping boundary.
* Note:            For better performance, implement this function as a macro:
*                  #define MACGet()    (EDATA)
*****************************************************************************
}
Public Inline Function MACGet() As EDATA
End Function

{
******************************************************************************
* Function:        WORD MACGetArray(BYTE *val, WORD len)
*
* PreCondition:    ERDPT must point to the place to read from.
* Input:           *val: Pointer to storage location
*                  len:  Number of bytes to read from the data buffer.
* Output:          Byte(s) of data read from the data buffer.
* Side Effects:    None
* Overview:        Reads several sequential bytes from the data buffer
*                  and places them into local memory.  ERDPT is incremented
*                  after each byte, following the same rules as MACGet().
* Note:            None
*****************************************************************************
}
Public Function MACGetArray(addr As Word, len As Word) As Word
    dim w as word
      
    w = len
    if (addr <> 0) then
        FSR2 = addr
        while (w <> 0)
            POSTINC2 = EDATA    // read data
            w = w - 1
        end while            
    else
        while (w <> 0)
            WREG = EDATA        // dummy read
            w = w - 1
        end while            
    endif
    
    result = len
End Function

{
******************************************************************************
* Function:        void MACPutArray(BYTE *val, WORD len)
*
* PreCondition:    EWRPT must point to the location to begin writing.
* Input:           *val: Pointer to source of bytes to copy.
*                  len:  Number of bytes to write to the data buffer.
* Output:          None
* Side Effects:    None
* Overview:        MACPutArray writes several sequential bytes to the
*                  Ethernet buffer RAM.  It performs faster than multiple MACPut()
*                  calls.  EWRPT is incremented by len.
* Note:            None
*****************************************************************************
}
Public Sub MACPutArray(addr As Word, len As Word)
    FSR2 = addr
    while (len <> 0)
        WRITE_EDATA(POSTINC2)
        len = len - 1
    end while
End Sub

inline function TBLRD_POSTINC() as TABLAT
    asm
        TBLRD*+
    end asm
end function

Public Sub MACPutROMArray(val As TROMAddress, len As Word)
    TABLEPTR = val
    while (len <> 0)
        TBLRD_POSTINC()
        WRITE_EDATA(TABLAT)
        len = len - 1
    end while
End Sub

{
******************************************************************************
* Function:        ReadPHYReg
*
* PreCondition:    Ethernet module must be enabled (ECON1.ETHEN = 1).
* Input:           Address of the PHY register to read from.
* Output:          16 bits of data read from the PHY register.
* Side Effects:    None
* Overview:        ReadPHYReg performs an MII read operation.  While in
*                  progress, it simply polls the MII BUSY bit wasting time
*                  (10.24us).
* Note:            None
*****************************************************************************
}
Function ReadPHYReg(reg As Byte) As PHYREG
   // Set the right address and start the register read operation
    MIREGADR = reg
    nop()
    MICMD = MICMD_MIIRD
    nop()

   // Loop to wait until the PHY register has been read through the MII
   // This requires 10.24us
    while (MISTAT.bits(BUSY) = 1) 
    end while

   // Stop reading
    MICMD = $00
    nop()
   
   // Obtain results and return
    result.Val.v(0) = MIRDL
    nop()
    result.Val.v(1) = MIRDH
End Function

{
******************************************************************************
* Function:        WritePHYReg
*
* PreCondition:    Ethernet Module must be enabled (ECON1.ETHEN = 1).
* Input:           Address of the PHY Register To write To.
*                   16 Bits of data To write To PHY Register.
* Output:          None
* Side Effects:    None
* Overview:        WritePHYReg performs an MII write operation.  While in
*                   progress, it simply polls the MII BUSY Bit wasting time
*                   (10.24us).
* Note:            None
*****************************************************************************
}
Sub WritePHYReg(reg as byte, data as word)
    dim t_intcon as byte
    
    // Write the register address
    MIREGADR = reg

    // Write the data through the MIIM interface
    // Order is important: write low byte first, high byte last
    //
    // Due to silicon Errata #5, you cannot access any register with LSb address
    // bits of 0x16 between your write to MIWRL and MIWRH or else the value in
    // MIWRL will be corrupted.  This inline assembly prevents this by copying
    // the value to PRODH:PRODL first, which is at fixed locations of
    // 0xFF4:0xFF3.  These addresses have LSb address bits of 0x14 and 0x13.
    // Interrupts must be disabled to prevent arbitrary ISR code from accessing
    // memory with LSb bits of 0x16 and corrupting the MIWRL value.

    // get a copy of the current INTCON and disable interrupts
    t_intcon = INTCON
    INTCON.bits(7) = 0

    PRODL = data.byte0
    PRODH = data.byte1 
    asm
        movff   PRODL, MIWRL
        nop
        movff   PRODH, MIWRH
    end asm
    
    // restore GIE (if it was set)
    if (t_intcon.bits(7) = 1) then
        INTCON.bits(7) = 1
    endif
    
    // Wait until the PHY register has been written
    // This operation requires 10.24us
    while (MISTAT.bits(BUSY) = 1)
    end while
 End Sub
 
{
******************************************************************************
* Macro:           void SetLEDConfig(Word NewConfig)
* PreCondition:    SPI bus must be initialized (done in MACInit()).
* Input:           NewConfig - 
*                        xxx0: Pulse stretching disabled
*                        xxx2: Pulse stretch To 40ms (default)
*                        xxx6: Pulse stretch To 73ms
*                        xxxA: Pulse stretch To 139ms
*                        
*                        xx1x: LEDB - TX
*                        xx2x: LEDB - RX (default)
*                        xx3x: LEDB - collisions
*                        xx4x: LEDB - link
*                        xx5x: LEDB - duplex
*                        xx7x: LEDB - TX And RX
*                        xx8x: LEDB - on
*                        xx9x: LEDB - off
*                        xxAx: LEDB - blink fast
*                        xxBx: LEDB - blink slow
*                        xxCx: LEDB - link And RX
*                        xxDx: LEDB - link And TX And RX
*                        xxEx: LEDB - duplex And collisions
*
*                        x1xx: LEDA - TX
*                        x2xx: LEDA - RX
*                        x3xx: LEDA - collisions
*                        x4xx: LEDA - link (default)
*                        x5xx: LEDA - duplex
*                        x7xx: LEDA - TX And RX
*                        x8xx: LEDA - on
*                        x9xx: LEDA - off
*                        xAxx: LEDA - blink fast
*                        xBxx: LEDA - blink slow
*                        xCxx: LEDA - link And RX
*                        xDxx: LEDA - link And TX And RX
*                        xExx: LEDA - duplex And collisions
* Output:          None
* Side Effects:    None
* Overview:        Writes the value of NewConfig into the PHLCON PHY register. 
*                  The LED pins will beginning outputting the new
*                  configuration immediately.
* Note:           
*****************************************************************************
}
Public Function GetLEDConfig() As PHYREG
   result = ReadPHYReg(PHLCON)
End Function

Public Sub SetLEDConfig(NewConfig As Word)
    // bits 15-12 are reserved, and must be written '0011'
    // bit 0 is also reserved, and should be written '0'
    NewConfig = (NewConfig and $0FFE) or $3000
    WritePHYReg(PHLCON, NewConfig)
End Sub

{
******************************************************************************
* Function:        void MACInit(void)
*
* PreCondition:    None
* Input:           None
* Output:          None
* Side Effects:    None
* Overview:        MACInit enables the Ethernet module, waits for the
*                  to become ready, and programs all registers for future
*                  TX/RX operations.
* Note:            This function blocks for at least 1ms, waiting for the
*                  hardware to stabilize.
*****************************************************************************
}
Public Sub MACInit()

    // Clear TRISA0 and TRISA1 to set LED0 and LED1 as outputs for Ethernet module status
    TRISA.bits(0) = 0   // Set LED0 as output
    TRISA.bits(1) = 0   // Set LED1 as output

    ECON2.bits(ETHEN) = 1   // enable ethernet (ETHEN)
    
    // wait for PHYRDY to become set
    while (ESTAT.bits(PHYRDY) = 0)
    end while
    
    // Configure the receive buffer boundary pointers
    // and the buffer write protect pointer (receive buffer read pointer)
    WasDiscarded = TRUE
    NextPacketLocation.Val = RXSTART
   
    ERXSTL = lowbyte(RXSTART)
    ERXSTH = highbyte(RXSTART)
   
    ERXRDPTL = lowbyte(RXSTOP)    // Write lo byte first
    ERXRDPTH = highbyte(RXSTOP) // Write  hi byte last
   
    ERXNDL = lowbyte(RXSTOP)
    ERXNDH = highbyte(RXSTOP)
   
    ETXSTL = lowbyte(TXSTART)
    ETXSTH = highbyte(TXSTART)
   
    // Write a permanant per packet control byte of $00
    EWRPTL = lowbyte(TXSTART)
    EWRPTH = highbyte(TXSTART)
    MACPut($00)

    // Configure Receive Filters
    // (No need to reconfigure - Unicast OR Broadcast with CRC checking is
    // acceptable)
    //ERXFCON = ERXFCON_CRCEN      // Promiscious mode

    // Configure the MAC
    // Enable the receive portion of the MAC
    MACON1 = MACON1_TXPAUS Or MACON1_RXPAUS Or MACON1_MARXEN 
    nop()

    // Pad packets to 60 bytes, add CRC, and check Type/Length field.
  #if (ETH_DUPLEX = FULL_DUPLEX)
    MACON3 = MACON3_PADCFG0 Or MACON3_TXCRCEN Or MACON3_FRMLNEN Or MACON3_FULDPX 
    nop()
    MABBIPG = $15 
    nop()
  #else
    MACON3 = MACON3_PADCFG0 Or MACON3_TXCRCEN Or MACON3_FRMLNEN 
    nop()
    MABBIPG = $12 
    nop()
  #endif

    // Allow infinite deferals if the medium is continuously busy
    // (do not time out a transmission if the half duplex medium is
    // completely saturated with other people's data)
    MACON4 = MACON4_DEFER 
    nop()

    // Set non-back-to-back inter-packet gap to 9.6us.  The back-to-back
    // inter-packet gap (MABBIPG) is set by MACSetDuplex() which is called
    // later
    MAIPGL = $12 
    nop()
    MAIPGH = $0C 
    nop()

    // Set the maximum packet size which the controller will accept
    MAMXFLL = lowbyte(6+6+2+1500+4) 
    nop()
    MAMXFLH = highbyte(6+6+2+1500+4) 
    nop()
   
    // Initialize physical MAC address registers
    MAADR1 = AppConfig.MyMACAddr.v(0) 
    nop()
    MAADR2 = AppConfig.MyMACAddr.v(1) 
    nop()
    MAADR3 = AppConfig.MyMACAddr.v(2) 
    nop()
    MAADR4 = AppConfig.MyMACAddr.v(3) 
    nop()
    MAADR5 = AppConfig.MyMACAddr.v(4) 
    nop()
    MAADR6 = AppConfig.MyMACAddr.v(5) 
    nop()

    // Disable half duplex loopback in PHY and set RXAPDIS bit as per errata
    WritePHYReg(PHCON2, PHCON2_HDLDIS Or PHCON2_RXAPDIS)

    // Configure LEDA to display LINK status, LEDB to display TX/RX activity
    SetLEDConfig($0472)
   
    // Set the PHY into the proper duplex state
  #if (ETH_DUPLEX = FULL_DUPLEX)
    WritePHYReg(PHCON1, PHCON1_PDPXMD)
  #else
    WritePHYReg(PHCON1, $0000)
  #endif

    // Enable packet reception
    ECON1.bits(RXEN) = 1
End Sub

{
******************************************************************************
* Function:        BOOL MACIsLinked()
*
* PreCondition:    None
* Input:           None
* Output:          TRUE: If the PHY reports that a link partner is present
*                        and the link has been up continuously since the last
*                        call to MACIsLinked()
*                  FALSE: If the PHY reports no link partner, or the link went
*                         down momentarily since the last call to MACIsLinked()
* Side Effects:    None
* Overview:        Returns the PHSTAT1.LLSTAT bit.
* Note:            None
*****************************************************************************
}
Public Function MACIsLinked() As Boolean
    Dim PHYRegister As PHYREG

    // LLSTAT is a latching low link status bit.  Therefore, if the link
    // goes down and comes back up before a higher level stack program calls
    // MACIsLinked(), MACIsLinked() will still return FALSE.  The next
    // call to MACIsLinked() will return TRUE (unless the link goes down
    // again).
    PHYRegister = ReadPHYReg(PHSTAT1)
    result = PHYRegister.PHSTAT1bits.LLSTAT
End Function

{
******************************************************************************
* Function:         MACFlush()
*
* PreCondition:    A packet has been created by calling MACPut() And
*                  MACPutHeader().
* Input:           None
* Output:          None
* Side Effects:    None
* Overview:        MACFlush causes the current TX packet To be sent out on
*               the Ethernet medium.  The hardware MAC will take control
*               And handle CRC generation, collision retransmission And
*               other details.
* Note:         After transmission completes (MACIsTxReady() returns TRUE),
*               the packet can be modified And transmitted again by calling
*               MACFlush() again.  Until MACPutHeader() Or MACPut() is
*               called (in the TX data area), the data in the TX buffer
*               will Not be corrupted.
*****************************************************************************
}
Public Sub MACFlush()
    // Reset the Ethernet TX logic.  This is a (suspected) errata workaround to
    // prevent the TXRTS bit from getting stuck set indefinitely, causing the
    // stack to lock up under certain bad conditions.
    ECON1.bits(TXRST) = 1       // TXRST
    ECON1.bits(TXRST) = 0    

    // Wait at least 1.6us after TX Reset before setting TXRTS.
    // If you don't wait long enough, the TX logic won't be finished resetting.
    delayus(2)   
    EIR.bits(TXERIF) = 0         // reset error flag (because of TXRST)

    // Start the transmission
    // After transmission completes (MACIsTxReady() returns TRUE), the packet
    // can be modified and transmitted again by calling MACFlush() again.
    // Until MACPutHeader() is called, the data in the TX buffer will not be
    // corrupted.
    ECON1.bits(TXRTS) = 1       // TXRTS
    wTXWatchdog = TickGet()     // record tx time for errata #7 fix
End Sub

{
******************************************************************************
* Function:        BOOL MACIsTxReady()
*
* PreCondition:    None
* Input:           None
* Output:          TRUE: If no Ethernet transmission is in progress
*                  FALSE: If a previous transmission was started, And it has
*                     Not completed yet.  While FALSE, the data in the
*                     transmit buffer And the TXST/TXND pointers must Not
*                     be changed.
* Side Effects:    None
* Overview:        Returns the ECON1.TXRTS Bit
* Note:            None
*****************************************************************************
}
Public Function MACIsTxReady() As Boolean
    if (ECON1.bits(TXRTS) = 0) then    // tx request to send (ECON1.3)
        result = true
        exit
    endif        

    // Retry transmission if the current packet seems to be not completing
    // Wait 3ms before triggering the retry (errata #7 tx deadlock fix)
    if((word(TickGet()) - wTXWatchdog) >= (3*TICK_SECOND/1000)) then
        ECON1.bits(TXRTS) = 0
        MACFlush()
    endif
    
    result = false
End Function

{
******************************************************************************
* Function:         MACDiscardRx()
*
* PreCondition:    None
* Input:           None
* Output:          None
* Side Effects:    None
* Overview:        Marks the last received packet (obtained using
*                  MACGetHeader())As being processed And frees the buffer
*                  memory associated With it
* Note:            It is safe To call this Function multiple times between
*                  MACGetHeader() calls.  Extra packets won't be thrown away
*                  Until MACGetHeader() makes it available.
*****************************************************************************
}
Public Sub MACDiscardRx()
    Dim NewRXRDLocation As WORD_VAL

    // Make sure the current packet was not already discarded
    If (WasDiscarded) Then
        Exit
    EndIf
    WasDiscarded = true
   
    // Decrement the next packet pointer before writing it into
    // the ERXRDPT registers.  This is a silicon errata workaround.
    // RX buffer wrapping must be taken into account if the
    // NextPacketLocation is precisely RXSTART.
    NewRXRDLocation.Val = NextPacketLocation.Val - 1
    if (NewRXRDLocation.Val < RXSTART) or (NewRXRDLocation.Val > RXSTOP) then
        NewRXRDLocation.Val = RXSTOP
    endif

    // Decrement the RX packet counter register, EPKTCNT
    ECON2.bits(PKTDEC) = 1       // PKTDEC

    // Move the receive read pointer to unwrite-protect the memory used by the
    // last packet.  The writing order is important: set the lo byte first,
    //  hi byte last.
    ERXRDPTL = NewRXRDLocation.v(0)
    ERXRDPTH = NewRXRDLocation.v(1)
 End Sub

{
******************************************************************************
* Function:        Word MACGetFreeRxSize()
*
* PreCondition:    None
* Input:           None
* Output:          A WORD estimate of how much RX buffer space is free at
*                  the present time.
* Side Effects:    None
* Overview:        None
* Note:            None
*****************************************************************************
}
Public Function MACGetFreeRxSize() As Word
    Dim ReadPT, WritePT As WORD_VAL
   
    // Read the Ethernet hardware buffer write pointer.  Because packets can be
    // received at any time, it can change between reading the low and high
    // bytes.  A loop is necessary to make certain a proper low/high byte pair
    // is read.
    Repeat
        // Save EPKTCNT in a temporary location
        ReadPT.v(0) = EPKTCNT
        WritePT.v(0) = ERXWRPTL
        WritePT.v(1) = ERXWRPTH
    Until (EPKTCNT = ReadPT.v(0))
   
    // Determine where the write protection pointer is
    ReadPT.v(0) = ERXRDPTL
    ReadPT.v(1) = ERXRDPTH
   
    // Calculate the difference between the pointers, taking care to account
    // for buffer wrapping conditions
    If (WritePT.Val > ReadPT.Val) Then
        result = (RXSTOP - RXSTART) - (WritePT.Val - ReadPT.Val)
    ElseIf (WritePT.Val = ReadPT.Val) Then
        result = RXSIZE - 1
    Else
        result = ReadPT.Val - WritePT.Val - 1
    EndIf
End Function

{
******************************************************************************
* Function:        BOOL MACGetHeader(MAC_ADDR *remote, Byte* Type)
*
* PreCondition:    None
*
* Input:           *remote: Location To store the Source MAC address of the
*                      received frame.
*               *Type: Location of a Byte To store the constant
*                     MAC_UNKNOWN, ETHER_IP, Or ETHER_ARP, representing
*                     the contents of the Ethernet Type field.
*
* Output:          TRUE: If a packet was waiting in the RX buffer.  The
*                    remote, And Type values are updated.
*               FALSE: If a packet was Not pending.  remote And Type are
*                     Not changed.
*
* Side Effects:    Last packet is discarded If MACDiscardRx() hasn't already
*               been called.
*
* Overview:        None
*
* Note:            None
*****************************************************************************
}
Public Function MACGetHeader(ByRef remote As MAC_ADDR, ByRef _Type As Byte) As Boolean
    Dim header As ENC_PREAMBLE

    // Test if at least one packet has been received and is waiting
    If (EPKTCNT = 0) Then
        result = false
        Exit
    EndIf

    // Make absolutely certain that any previous packet was discarded
    If (WasDiscarded = false) Then
        MACDiscardRx()
        result = false
        Exit
    EndIf

    // Save the location of this packet
    CurrentPacketLocation.Val = NextPacketLocation.Val

    // Set the read pointer to the beginning of the next unprocessed packet
    ERDPTL = CurrentPacketLocation.v(0)
    ERDPTH = CurrentPacketLocation.v(1)

    // Obtain the MAC header from the Ethernet buffer
    MACGetArray(@header, SizeOf(header))

    // The EtherType field, like most items transmitted on the Ethernet medium
    // are in big endian.
    header._Type.Val = NETUtils.swaps(header._Type.Val)

    // Do a sanity check.  There might be a bug in code someplace if this
    // Reset() ever happens.  Check for potential errors in array/pointer writing code.
   If (header.NextPacketPointer > RXSTOP) Or
      (header.NextPacketPointer.bits(0) = 1) Or
      (header.StatusVector.Bits.zero = 1) Or
      (header.StatusVector.Bits.CRCError = 1) Or
      (header.StatusVector.Bits.ByteCount > 1518) Or
      not header.StatusVector.Bits.ReceiveOk Then
        Reset()
    EndIf

    // Save the location where the hardware will write the next packet to
    NextPacketLocation.Val = header.NextPacketPointer

    // Return the Ethernet frame's Source MAC address field to the caller
    // This parameter is useful for replying to requests without requiring an
    // ARP cycle.
    // memcpy((void*)remote->v, (void*)header.SourceMACAddr.v, SizeOf( *remote));
    remote.v = header.SourceMACAddr.v

    // Return a simplified version of the EtherType field to the caller
    _Type = MAC_UNKNOWN
    If (header._Type.v(1) = $08) And (header._Type.v(0) = ETHER_IP Or header._Type.v(0) = ETHER_ARP) Then
        _Type = header._Type.v(0)
    EndIf

    // Mark this packet as discardable
    WasDiscarded = false
    result = true
End Function

{
******************************************************************************
* Function:         MACPutHeader(MAC_ADDR *remote, BYTE type, WORD dataLen)
*
* PreCondition:    MACIsTxReady() must return TRUE.
* Input:           *remote: Pointer to memory which contains the destination
*                       MAC address (6 bytes)
*                   type: The constant ETHER_ARP or ETHER_IP, defining which
*                    value to write into the Ethernet header's type field.
*                   dataLen: Length of the Ethernet data payload
* Output:          None
* Side Effects:    None
* Overview:        None
* Note:            Because of the dataLen parameter, it is probably
*               advantageous to call this function immediately before
*               transmitting a packet rather than initially when the
*               packet is first created.  The order in which the packet
*               is constructed (header first or data first) is not
*               important.
*****************************************************************************
} 
Public Sub MACPutHeader(ByRef remote As MAC_ADDR, _Type As Byte, dataLen As Word)
    // Set the write pointer to the beginning of the transmit buffer
    EWRPTL = lowbyte(TXSTART+1)
    EWRPTH = highbyte(TXSTART+1)

    // Calculate where to put the TXND pointer
    dataLen = dataLen + SizeOf(MAC_ETHER_HEADER) + TXSTART

    // Write the TXND pointer into the registers, given the dataLen given
    ETXNDL = dataLen.Byte0
    ETXNDH = dataLen.Byte1
   
    // Set the per-packet control byte and write the Ethernet destination
    // address
    MACPutArray(@remote, SizeOf(MAC_ADDR))

    // Write our MAC address in the Ethernet source field
    MACPutArray(@AppConfig.MyMACAddr, SizeOf(AppConfig.MyMACAddr))

    // Write the appropriate Ethernet Type WORD for the protocol being used
    MACPut($08)
    If (_Type = MAC_IP) Then
        MACPut(ETHER_IP)
    Else
        MACPut(ETHER_ARP)
    EndIf
End Sub

{
******************************************************************************
* Function:         MACSetReadPtrInRx(Word offset)
*
* PreCondition:     A packet has been obtained by calling MACGetHeader() And
*                   getting a TRUE result.
* Input:            offset: Word specifying how many Bytes beyond the Ethernet
*                   header's type field to relocate the SPI read
*                   pointer.
* Output:          None
* Side Effects:    None
* Overview:        SPI Read pointer are updated.  All calls To
*                   MACGet() And MACGetArray() will use these new values.
* Note:         RXSTOP must be statically defined As being > RXSTART For
*               this Function To work correctly.  In other Words, do Not
*               define an RX buffer which spans the $1FFF->$0000 memory
*               boundary.
*****************************************************************************
}
Public Sub MACSetReadPtrInRx(offset As Word )
    Dim ReadPT As WORD_VAL 

    // Determine the address of the beginning of the entire packet
    // and adjust the address to the desired location
    ReadPT.Val = CurrentPacketLocation.Val + SizeOf(ENC_PREAMBLE) + offset
   
    // Since the receive buffer is circular, adjust if a wraparound is needed
    If (ReadPT.Val > RXSTOP) Then
        ReadPT.Val = ReadPT.Val - RXSIZE
    EndIf
   
    // Set the read pointer to the new calculated value
    ERDPTL = ReadPT.v(0)
    ERDPTH = ReadPT.v(1)
End Sub

{
******************************************************************************
* Function:        Word MACSetWritePtr(Word Address)
*
* PreCondition:    None
* Input:           Address: Address To seek To
* Output:          Word: Old EWRPT location
* Side Effects:    None
* Overview:        SPI Write pointer is updated.  All calls To
*                   MACPut() And MACPutArray() will use this new value.
* Note:         None
*****************************************************************************
}
Public Function MACSetWritePtr(address As Word) As Word
    // set return val to current 
    result.byte0 = EWRPTL
    result.byte1 = EWRPTH

    // Set the write pointer to the new calculated value
    EWRPTL = address.Byte0
    EWRPTH = address.Byte1
End Function

{
******************************************************************************
* Function:        Word MACSetReadPtr(Word Address)
*
* PreCondition:    None
* Input:           Address: Address To seek To
* Output:          Word: Old ERDPT value
* Side Effects:    None
* Overview:        SPI Write pointer is updated.  All calls To
*                  MACPut() And MACPutArray() will use this new value.
* Note:         None
*****************************************************************************
}
Public Function MACSetReadPtr(address As Word) As Word
    // set return val to current 
    result.byte0 = ERDPTL
    result.byte1 = ERDPTH

    // Set the write pointer to the new calculated value
    ERDPTL = address.Byte0
    ERDPTH = address.Byte1
End Function

{
******************************************************************************
* Function:        Word CalcIPBufferChecksum(Word len)
*
* PreCondition:    Read buffer pointer set To starting of checksum data
* Input:           len: Total number of Bytes To calculate the checksum over.
*                   The first Byte included in the checksum is the Byte
*                   pointed To by ERDPT, which is updated by calls To
*                   MACSetReadPtr(), MACGet(), MACGetArray(),
*                   MACGetHeader(), etc.
* Output:          16-Bit checksum As defined by RFC 793
* Side Effects:    None
* Overview:        This Function performs a checksum calculation in the MAC
*                  buffer itself
* Note:            This Function works either in the RX buffer area Or the TX
*                  buffer area.  No validation is Done on the len parameter.
*****************************************************************************
}
Public Function CalcIPBufferChecksum(len As Word) As Word
    Dim wStart As word
    Dim Checksum As DWORD_VAL
    Dim ChunkLen as word
    Dim wTemp as word
    Dim DataBuffer(20) As Byte   // Must be an even size
    Dim DataPtr As FSR0
   
    Checksum.Val = $00000000
   
    // Save the read pointer starting address
    wStart.byte0 = ERDPTL
    wStart.byte1 = ERDPTH

    While (len <> 0)
        // Obtain a chunk of data (less SPI overhead compared
        // to requesting one byte at a time)
        If (len > SizeOf(DataBuffer)) Then
            ChunkLen = SizeOf(DataBuffer)
        Else
            ChunkLen = len
        EndIf
         
        MACGetArray(@DataBuffer, ChunkLen)

        len = len - ChunkLen

        // Take care of a last odd numbered data byte
        If (ChunkLen.bits(0) = 1) Then
            DataBuffer(ChunkLen) = $00
            ChunkLen = ChunkLen + 1
        EndIf

        // Calculate the checksum over this chunk
        DataPtr = @DataBuffer
        While (ChunkLen <> 0)
            wTemp.Byte0 = POSTINC0
            wTemp.Byte1 = POSTINC0
            Checksum.Val = Checksum.Val + wTemp
            ChunkLen = ChunkLen - 2
        End While
    End While
   
    // Restore old read pointer location
    ERDPTL = wStart.byte0
    ERDPTH = wStart.byte1
   
    // Do an end-around carry (one's complement arrithmatic)
    Checksum.Val = Checksum.w(0) + Checksum.w(1)

    // Do another end-around carry in case if the prior add
    // caused a carry out
    Checksum.w(0) = Checksum.w(0) + Checksum.w(1)

    // Return the resulting checksum
    result = Not Checksum.w(0)
End Function

{
******************************************************************************
* Function:        Word MACCalcRxChecksum(Word offset, Word len)
*
* PreCondition:    None
* Input:           offset   - Number of Bytes beyond the beginning of the
*                     Ethernet data (first Byte after the Type field)
*                     where the checksum should begin
*                   len      - Total number of Bytes To Include in the checksum
* Output:          16-Bit checksum As defined by RFC 793.
* Side Effects:    None
* Overview:        This Function performs a checksum calculation in the MAC
*                  buffer itself
* Note:            None
*****************************************************************************
}
Public Function MACCalcRxChecksum(offset As Word, len As Word) As Word
    Dim wTemp As word
    Dim wRDSave As word

   // Add the offset requested by firmware plus the Ethernet header
   wTemp = CurrentPacketLocation.Val + SizeOf(ENC_PREAMBLE) + offset
   If (wTemp > RXSTOP) Then      // Adjust value if a wrap is needed
      wTemp = wTemp - RXSIZE
   EndIf
   
   wRDSave.byte0 = ERDPTL
   wRDSave.byte1 = ERDPTH

   ERDPTL = wTemp.byte0
   ERDPTH = wTemp.byte1
   
   wTemp = CalcIPBufferChecksum(len)

   ERDPTL = wRDSave.byte0
   ERDPTH = wRDSave.byte1
   result = wTemp
End Function
 
{
******************************************************************************
* Function:         MACMemCopyAsync(Word destAddr, Word sourceAddr, Word len)
*
* PreCondition:    None
* Input:           destAddr:   Destination address in the Ethernet memory To
*                        copy To.  If the MSb is set, the current EWRPT
*                        value will be used instead.
*                   sourceAddr:   Source address To Read from.  If the MSb is
*                        set, the current ERDPT value will be used
*                        instead.
*                   len:      Number of Bytes To copy
* Output:          None
* Side Effects:    None
* Overview:        Bytes are asynchrnously transfered within the buffer.  Call
*                   MACIsMemCopyDone() To see when the transfer is complete. 
* Note:            If a prior transfer is already in progress prior To
*                   calling this Function, this Function will block Until it
*                   can start this transfer.
*****************************************************************************
} 
 Public Sub MACMemCopyAsync(destAddr As Word, sourceAddr As Word, len As Word)
    Dim ReadSave, WriteSave As WORD_VAL
    Dim UpdateWritePointer As Boolean
    Dim UpdateReadPointer As Boolean

    UpdateWritePointer = false
    UpdateReadPointer = false
 
    If (destAddr.bits(15) = 1) Then
        UpdateWritePointer = true
        destAddr.Byte0 = EWRPTL
        destAddr.Byte1 = EWRPTH
    EndIf   

    If (sourceAddr.bits(15) = 1) Then
        UpdateReadPointer = true
        sourceAddr.Byte0 = ERDPTL
        sourceAddr.Byte1 = ERDPTH
    EndIf

    // Handle special conditions where len == 0 or len == 1
    // The DMA module is not capable of handling those corner cases
    If (len <= 1) Then
        ReadSave.v(0) = ERDPTL
        ReadSave.v(1) = ERDPTH
        WriteSave.v(0) = EWRPTL
        WriteSave.v(1) = EWRPTH
        ERDPTL = sourceAddr.Byte0
        ERDPTH = sourceAddr.Byte1
        EWRPTL = destAddr.Byte0
        EWRPTH = destAddr.Byte1
        While (len <> 0)
            MACPut(MACGet())
            len = len - 1
        end while
        If (Not UpdateReadPointer) Then
            ERDPTL =  ReadSave.v(0)
            ERDPTH =  ReadSave.v(1)
        EndIf
        If (Not UpdateWritePointer) Then
            EWRPTL = WriteSave.v(0)
            EWRPTH = WriteSave.v(1)
        EndIf
    Else
        If (UpdateWritePointer) Then
            WriteSave.Val = destAddr + len
            EWRPTL = WriteSave.v(0)
            EWRPTH = WriteSave.v(1)
        EndIf
        len = len + (sourceAddr - 1)
        While (ECON1.bits(DMAST) = 1)
        end while
        EDMASTL = sourceAddr.Byte0
        EDMASTH = sourceAddr.Byte1
        EDMADSTL = destAddr.Byte0
        EDMADSTH = destAddr.Byte1
   
        If (sourceAddr <= RXSTOP) And (len > RXSTOP) Then //&& (sourceAddr >= RXSTART))
            len = len - RXSIZE
        EndIf
         
        EDMANDL = len.Byte0
        EDMANDH = len.Byte1
    
        ECON1.bits(CSUMEN) = 0       // CSUMEN
        ECON1.bits(DMAST) = 1       // DMAST
        While (ECON1.bits(DMAST) = 1)           // DMA requires that you must not access EDATA while DMA active
        end while

        If (UpdateReadPointer) Then
            len = len + 1
            If (sourceAddr <= RXSTOP) And (len > RXSTOP) Then //&& (sourceAddr >= RXSTART))
                len = len - RXSIZE
            EndIf   
            ERDPTL = len.Byte0
            ERDPTH = len.Byte1
        EndIf
    EndIf
End Sub

Public Function MACIsMemCopyDone() As Boolean
    result = false
    if (ECON1.bits(DMAST) = 0) then
        result = true
    endif
End Function   

{
******************************************************************************
* Function:         MACPowerDown()
*
* PreCondition:    None
* Input:           None
* Output:          None
* Side Effects:    None
* Overview:        MACPowerDown disables the Ethernet Module.
*                   All MAC And PHY registers should Not be accessed.
* Note:            Normally, this Function would be called before putting the
*                   PIC To sleep.  If a packet is being transmitted While this
*                   Function is called, this Function will block Until it is
*                   it complete. If anything is being received, it will be
*                   completed.
*               
*                   The Ethernet Module will Continue To draw significant
*                   power in sleep mode If this Function is Not called first.
*****************************************************************************
}
Sub MACPowerDown()
   // Disable packet reception
   ECON1.bits(RXEN) = 0

   // Make sure any last packet which was in-progress when RXEN was cleared
   // is completed
   While (ESTAT.bits(RXBUSY) = 1)
   End While

   // If a packet is being transmitted, wait for it to finish
   While (ECON1.bits(TXRTS) = 1)
   End While
   
   // Disable the Ethernet module
   ECON2.bits(ETHEN) = 0
End Sub

{
******************************************************************************
* Function:         MACPowerUp()
*
* PreCondition:    None
* Input:           None
* Output:          None
* Side Effects:    None
* Overview:        MACPowerUp returns the Ethernet Module back To normal operation
*               after a previous call To MACPowerDown().  Calling this
*               Function when already powered up will have no effect.
* Note:            If a link partner is present, it will take 10s of
*               milliseconds before a new link will be established after
*               waking up.  While Not linked, packets which are
*               transmitted will most likely be lost.  MACIsLinked() can
*               be called To determine If a link is established.
*****************************************************************************
}
Sub MACPowerUp()
   // Power up the Ethernet module
   ECON2.bits(ETHEN) = 1

   // Wait for PHY to become ready
   While (ESTAT.bits(PHYRDY) = 0)
   End While

   // Enable packet reception
   ECON1.bits(RXEN) = 1
End Sub

//
// module initialization
//
wTXWatchdog = 0

End Module
