{
*****************************************************************************
*  Name    : NETApp.BAS                                                     *
*  Author  : David John Barker                                              *
*  Notice  : Copyright (c) 2007 Mecanique                                   *
*          : All Rights Reserved                                            *
*  Date    : 09/10/2007                                                     *
*  Version : 1.0                                                            *
*  Notes   :                                                                *
*          :                                                                *
*****************************************************************************
}
Module NETApp


// stack related modules...
Include "NETConfig.bas"
Include "NETTypes.bas"

Include "NBNS.bas"
Include "DHCP.bas"
Include "SMTP.bas"
Include "SNTP.bas"
Include "StackTsk.bas"
Include "Tick.bas"

(*
   The microchip MAC Vendor ID is 00-04-A3, which has been registered
   with the IEEE Standards OUI database. You really should keep this
   and assign a unique ID to the last three bytes of the MAC address
*)

// default is configure from EEPROM, if the MCU supports it...
#if _eeprom <> 0
#option NET_ROM_CONFIG = false
#else
#option NET_ROM_CONFIG = true
#endif
   
// configuration details held in EEPROM...
#if Not NET_ROM_CONFIG
Include "EEPROM.bas"
EEPROM(@EEMACAddr)              = ($00, $04, $A3, $00, $00, $00)  // 6 byte MAC address
EEPROM(@EEDefaultIPAddr)        = (192, 168, 254, 100)            // IP address
EEPROM(@EEDefaultHostName)      = ("swordfish", 0)                // NetBIOS name

// if DHCP used, these values can all be zero...
EEPROM(@EEDefaultMask)          = (255, 255, 255, 0)             // Mask
EEPROM(@EEDefaultGateway)       = (0, 0, 0, 0)                   // Gateway
EEPROM(@EEPrimaryDNS)           = (0, 0, 0, 0)                   // Primary DNS
EEPROM(@EESecondaryDNS)         = (0, 0, 0, 0)                   // Secondary DNS

// configuration details - if not found in EEPROM...
#else
Public Const
   MY_DEFAULT_MAC(6) As Byte           = ($00, $04, $A3, $00, $00, $00),
   MY_DEFAULT_IP_ADDR(4) As Byte       = (192, 168, 254, 100),
   MY_DEFAULT_HOST_NAME                = "swordfish",

   // if DHCP used, these values can all be zero...
   MY_DEFAULT_MASK(4) As Byte          = (255, 255, 255, 0),
   MY_DEFAULT_GATE(4) As Byte          = (0, 0, 0, 0),
   MY_DEFAULT_PRIMARY_DNS(4) As Byte   = (0, 0, 0, 0),
   MY_DEFAULT_SECONDARY_DNS(4) As Byte = (0, 0, 0, 0)
#endif
   
// helper modules...
Include "String.bas"

#if NET_DHCP
Dim DHCPBindCount As Byte
#endif
Dim TaskEnabled As Boolean
{
****************************************************************************
* Name    : DHCPEvent                                                      *
* Purpose : Returns true if a DHCP event has occurred, false otherwise     *
****************************************************************************
}
Public Function DHCPEvent() As Boolean
   result = false
   #if NET_DHCP
   If AppConfig.Flags.bIsDHCPEnabled Then
      If DHCPBindCount <> DHCP.DHCPBindCount Then
         DHCPBindCount = DHCP.DHCPBindCount      
         TaskEnabled = true
         result = true         
      EndIf   
   EndIf
   #endif
End Function
{
****************************************************************************
* Name    : Task                                                           *
* Purpose : Performs stack processing tasks                                *
****************************************************************************
}
Public Sub Task()
   StackTask()
   
   If TaskEnabled Then
      // NETBIOS task...
      #if NET_NBNS
      NBNSTask()
      #endif 
  
      //  mail task...
      #if NET_SMTP
      SMTPTask()  
      #endif 
   
      // time server task...
      #if NET_SNTP
      SNTPClient()
      #endif
   EndIf   
End Sub
{
****************************************************************************
* Name    : ConfigureHostname (PRIVATE)                                    *
* Purpose : Ensures a given NETBIOS hostname is 16 characters long         *
****************************************************************************
}
Sub ConfigureHostname(pHostName As String)
   Dim Index As Byte
   
   FSR1 = @pHostName                 // source
   FSR0 = @AppConfig.NetBIOSName     // destination
   Index = 0
   
   // copy host name to app config host name - host name is
   // 0..14 characters, plus null terminator (total is 16 bytes)
   While INDF1 <> 0 And Index < SizeOf(AppConfig.NetBIOSName) - 1
      POSTINC0 = Uppercase(POSTINC1)
      Inc(Index)
   Wend   
   
   // any remaining characters, pack with space...
   While Index < SizeOf(AppConfig.NetBIOSName) - 1
      POSTINC0 = Byte(" ")
      Inc(Index)
   Wend
   
   // last char is a null terminator...
   POSTINC0 = 0
End Sub
{
****************************************************************************
* Name    : SetHostname (PRIVATE)                                          *
* Purpose : Set NET application hostname                                   *
****************************************************************************
}
#if Not NET_ROM_CONFIG
Sub SetHostname()
   Dim Hostname As String
   Dim Value, Addr As Byte
   FSR0 = @Hostname
   Addr = EEDefaultHostName
   Repeat
      Value = EE.ReadByte(Addr)
      POSTINC0 = Value
      Inc(Addr)
   Until Value = 0   
   ConfigureHostname(Hostname)
End Sub
{
****************************************************************************
* Name    : AppInit (PRIVATE)                                              *
* Purpose : Configures MAC, IP, Mask and Gateway addresses (EEPROM)        *
****************************************************************************
}
Sub AppInit()
   Dim Addr, Index As Byte
   
   // load MAC address from EEPROM...
   FSR0 = @AppConfig.MyMACAddr             // MAC address
   Addr = EEMACAddr                        // EEPROM address
   For Index = 1 To 6                      // read 6 bytes
      POSTINC0 = EE.ReadByte(Addr)  
      Inc(Addr)
   Next
      
   // load IP address from EEPROM...
   AppConfig.MyIPAddr.Val = EE.ReadLongWord(EEDefaultIPAddr)
   AppConfig.DefaultIPAddr.Val = AppConfig.MyIPAddr.Val
   
   // load mask from EEPROM...
   AppConfig.MyMask.Val = EE.ReadLongWord(EEDefaultMask)
   AppConfig.DefaultMask.Val = AppConfig.MyMask.Val
  
   // load gateway...
   AppConfig.MyGateway.Val = EE.ReadLongWord(EEDefaultGateway)

   // load primary DNS and secondary DNS from EEPROM...
   #if defined(STACK_USE_DNS)
   AppConfig.PrimaryDNSServer.Val = EE.ReadLongWord(EEPrimaryDNS)
   AppConfig.SecondaryDNSServer.Val = EE.ReadLongWord(EESecondaryDNS)
   #endif
   
   // load the hostname from EEPROM (NetBIOS name)
   #if NET_NBNS
   SetHostname
   #endif   
End Sub
#endif
{
****************************************************************************
* Name    : AppInit (PRIVATE)                                              *
* Purpose : Configures MAC, IP, Mask and Gateway addresses (ROM)           *
****************************************************************************
}
#if NET_ROM_CONFIG
Sub AppInit()
   AppConfig.MyMACAddr.v = MY_DEFAULT_MAC
   AppConfig.MyIPAddr.v = MY_DEFAULT_IP_ADDR
   AppConfig.MyMask.v = MY_DEFAULT_MASK
   AppConfig.MyGateway.v = MY_DEFAULT_GATE
   AppConfig.DefaultIPAddr.Val = AppConfig.MyIPAddr.Val
   AppConfig.DefaultMask.Val = AppConfig.MyMask.Val
   #if defined(STACK_USE_DNS)
   AppConfig.PrimaryDNSServer.v = MY_DEFAULT_PRIMARY_DNS
   AppConfig.SecondaryDNSServer.v = MY_DEFAULT_SECONDARY_DNS
   #endif
   #if NET_NBNS
   ConfigureHostname(MY_DEFAULT_HOST_NAME)
   #endif
End Sub
#endif

// initialize
#if NET_DHCP
DHCPBindCount = DHCP.DHCPBindCount
TaskEnabled = false
#else
TaskEnabled = true
#endif

AppInit()          // initialize NET application
TickInit()         // initialize NET timer
StackInit()        // Initialize core stack layers (MAC, ARP, TCP, UDP etc)





