program EthernetTCPServer

//
//-----------------------------------------------------------------------------
// hardware platforms (uncomment one section/customize for your hardware)
// the USER_INIT file for each platform contains the pic config settings, 
// userInit() routine, LED definitions, and any misc setup info
//-----------------------------------------------------------------------------
//
'{
// LJCV Electronics eIP-10 18F2620 with ENC28J60 (http://www.ljcv.net/eip10/)
device = 18F2620
clock = 40
#option USER_INIT = "EIP10_2620.bas"
#option ETH_MAC = ENC28J60
'}

{
// Microchip explorer18 (18F8722) with pictail plus (ETH_MAC = ENCx24J600)
device = 18F8722
clock = 32
#option USER_INIT = "EXP18_8722.bas"
#option ETH_MAC = ENCx24J600
}

{
// Microchip explorer18 with 18F87K22 PIM and pictail plus (ETH_MAC = ENCx24J600)
device = 18F87K22
clock = 64
#option USER_INIT = "EXP18_87K22.bas"
#option ETH_MAC = ENCx24J600
}

{
// Olimex PIC-WEB board 18F67J60 (https://www.olimex.com/Products/PIC/Development/PIC-WEB/)
device = 18F67J60
// select clock - either 25MHz or 41.6667MHz
'clock = 25      // default
clock = 41.6667    // 5xPLL=enabled, PLL Prescaler=div3, PLL Postscaler=disabled
#option USER_INIT = "PIC_WEB_67J60.bas"
'#option ETH_MAC = ETHXXJ60
}

//
//-----------------------------------------------------------------------------
// hardware setup
//-----------------------------------------------------------------------------
//
// include the board support USER_INIT file specified above
include USER_INIT

// USART support
// note: recommend using the modified USART.BAS module v1.6
#option USART_BRGH = true
#option USART_BRG16 = true
Include "usart.bas"

// for the ENC28J60 or ETHXXJ60, the default duplex setting is HALF_DUPLEX
// FULL_DUPLEX may or may not work depending on your network hardware setup 
// (since they don't support auto-negotiation). The ENCx24J600 will negotiate
// 10/100 speed and duplex settings
'#option ETH_DUPLEX = FULL_DUPLEX

// important configurations...
#option NET_ICMP = true   // enable ping 
#option NET_DHCP = true   // enable Dynamic Host Configuration Protocol (DHCP)
#option NET_NBNS = true   // enable NETBIOS
#option NET_TCP = true    // enable TCP
Include "NETApp.bas"      // always first!

// TCPIP stack specifics...
Include "NETTypes.bas"
Include "NETUtils.bas"
Include "Tick.bas"
Include "TCP.bas"

// program helpers...
Include "NETAppDisplay.bas"
Include "convert.bas"

// maximum number of TCP server sockets
Const MaxSockets = MAX_TCP_SOCKETS // found in NETConfig, default is eight 

// state machine constants...
Const
    SM_HOME = 0,
    SM_LISTENING = 1

// structure that holds socket state, connected flag and TCP socket ID...
Structure TSocket
    State As Byte
    Connected As Boolean
    ID As TCP_SOCKET
    dummy As Byte // 4 byte align
End Structure

// TCP server socket array...
Dim Sock(MaxSockets) As TSocket

{
****************************************************************************
* Name    : TCPServerInit                                                  *
* Purpose : Prepare socket array                                           *
****************************************************************************
}
Sub TCPServerInit()
    Dim Index As Byte
    Dim Socket As TSocket
   
    For Index = 0 To MaxSockets - 1
        Socket.ID = INVALID_SOCKET
        Socket.Connected = false
        Socket.State = SM_HOME
        Sock(index) = Socket
    Next
End Sub

{
****************************************************************************
* Name    : TCPServerOpen                                                  *
* Purpose : Open for listening 'MaxSockets' TCP sockets                    *
****************************************************************************
}
Sub TCPServerOpen()
    Const SERVER_PORT = 9760
   
    Dim Index As Byte
    Dim Socket As TSocket
    For Index = 0 To MaxSockets - 1
        Socket = Sock(index)
        If Socket.ID = INVALID_SOCKET Then
            Socket.ID = TCPListen(SERVER_PORT)
            If Socket.ID <> INVALID_SOCKET Then
                Socket.State = SM_LISTENING
            EndIf
        EndIf
        Sock(index) = Socket
    Next
End Sub

{
****************************************************************************
* Name    : TCPServerTask                                                  *
* Purpose : Process the TCP server sockets                                 *
****************************************************************************
}
Sub TCPServerTask()
    Dim c As Byte
    Dim Index As Byte
    Dim Socket As TSocket

    For Index = 0 To MaxSockets - 1 
        Socket = Sock(index)

        // server socket allocated...
        If Socket.State = SM_LISTENING Then

            // see if a client has disconnected...
            If Not TCPIsConnected(Socket.ID) Then
                If Socket.Connected Then
                    Socket.Connected = false
                    USART.Write("Client ", DecToStr(Socket.ID), " disconnected",13,10)
                EndIf
                Sock(index) = Socket
                Continue
            EndIf

            // see if a client has just connected...
            If Not Socket.Connected Then
                Socket.Connected = true
                USART.Write("Client connected, ")
                TCPGetRemoteInfo(Socket.ID)
                USART.Write("IP address is ")
                DisplayIPAddress(TCP.RemoteInfo.Remote.IPAddr)
            EndIf

            // receive and display any bytes sent...
            If TCPIsGetReady(Socket.ID) <> 0 Then
                While TCPGet(Socket.ID, c)
                    USART.Write(c)
                Wend
                USART.Write(13,10)

                // transmit  a reply...
                If TCPIsPutReady(Socket.ID) <> 0 Then
                    TCPPutString(Socket.ID,"Message from server : " + DecToStr(TickGet))
                EndIf
            EndIf
        EndIf
        Sock(index) = Socket
    Next
End Sub

// program start...
usart.SetBaudrate(br115200)
usart.write(13, 10, "**TCP server test - run TCPClient.exe tool**", 13, 10)
TCPServerInit()

#if NET_DHCP
USART.Write("Waiting for DHCP...",13,10)
#else
TCPServerOpen()
#endif

// loop forever...
While true
   // This task performs normal stack task including checking for incoming 
   // packet, type of packet and calling appropriate stack entity to process it.
   NETApp.Task
   TCPServerTask() 

   // If DHCP is enabled, display configuration when a DHCP event
   // has occurred - also open our TCP server sockets for listening...
   #if NET_DHCP
   If NETApp.DHCPEvent() Then
      DisplayConfig()
      TCPServerOpen()
   EndIf
   #endif
Wend

end program
