program EthernetICMP_noDNS

//
//-----------------------------------------------------------------------------
// hardware platforms (uncomment one section/customize for your hardware)
// the USER_INIT file for each platform contains the pic config settings, 
// userInit() routine, LED definitions, and any misc setup info
//-----------------------------------------------------------------------------
//
'{
// LJCV Electronics eIP-10 18F2620 with ENC28J60 (http://www.ljcv.net/eip10/)
device = 18F2620
clock = 40
#option USER_INIT = "EIP10_2620.bas"
#option ETH_MAC = ENC28J60
'}

{
// Microchip explorer18 (18F8722) with pictail plus (ETH_MAC = ENCx24J600)
device = 18F8722
clock = 32
#option USER_INIT = "EXP18_8722.bas"
#option ETH_MAC = ENCx24J600
}

{
// Microchip explorer18 with 18F87K22 PIM and pictail plus (ETH_MAC = ENCx24J600)
device = 18F87K22
clock = 64
#option USER_INIT = "EXP18_87K22.bas"
#option ETH_MAC = ENCx24J600
}

{
// Olimex PIC-WEB board 18F67J60 (https://www.olimex.com/Products/PIC/Development/PIC-WEB/)
device = 18F67J60
// select clock - either 25MHz or 41.6667MHz
'clock = 25      // default
clock = 41.6667    // 5xPLL=enabled, PLL Prescaler=div3, PLL Postscaler=disabled
#option USER_INIT = "PIC_WEB_67J60.bas"
'#option ETH_MAC = ETHXXJ60
}

//
//-----------------------------------------------------------------------------
// hardware setup
//-----------------------------------------------------------------------------
//
// include the board support USER_INIT file specified above
include USER_INIT

// USART support
// note: recommend using the modified USART.BAS module v1.6
#option USART_BRGH = true
#option USART_BRG16 = true
Include "usart.bas"

// for the ENC28J60 or ETHXXJ60, the default duplex setting is HALF_DUPLEX
// FULL_DUPLEX may or may not work depending on your network hardware setup 
// (since they don't support auto-negotiation). The ENCx24J600 will negotiate
// 10/100 speed and duplex settings
'#option ETH_DUPLEX = FULL_DUPLEX

// important configurations...
#option NET_TICK_PRIORITY = ipLow       // set low priority tick (serial isrrx defaults to high)
#option NET_ICMP = true   // enable ping 
#option NET_DHCP = true   // enable Dynamic Host Configuration Protocol (DHCP)
#option NET_NBNS = true   // enable NETBIOS
Include "NETApp.bas"      // always first!

// TCPIP stack specifics...
Include "NETTypes.bas"
Include "NETUtils.bas"
Include "Tick.bas"
Include "ICMP.bas"

// program helpers...
Include "NETAppDisplay.bas"
Include "convert.bas"
Include "ISRRX.bas"

// state machine constants...
Const
    SM_HOME = 0,
    SM_GET_RESPONSE = 1,
    SM_FINISHED = 2

// ping variables...   
Dim 
    PingState As Byte,
    PingTimer As TICK

{
****************************************************************************
* Name    : Init                                                           *
* Purpose : Initialise timer and state machine                             *
****************************************************************************
}
Sub Init()
    PingTimer = 0
    PingState = SM_HOME
End Sub

{
****************************************************************************
* Name    : Ping                                                           *
* Purpose : Start a ping                                                   *
****************************************************************************
}
Sub Ping(ByRef pIP As String)
    Dim PingIP As IP_ADDR

    If PingState = SM_HOME Then
        If ICMPBeginUsage() Then
            NETUtils.StringToIPAddress(@pIP, PingIP)
            ICMPSendPing(PingIP.Val)
            PingState = SM_GET_RESPONSE
            PingTimer = TickGet()
        EndIf
    EndIf
End Sub

{
****************************************************************************
* Name    : PingTask                                                       *
* Purpose : Process a ping task                                            *
****************************************************************************
}
Function PingTask() As SHORT
    Result = -2 // do nothing...

    // ping has been started, now wait for ping response...
    If PingState = SM_GET_RESPONSE Then
        Result = ICMPGetReply()
        // the request timed out...
        If Result <> - 2 Then
            PingState = SM_FINISHED
        EndIf
    EndIf

    // wait at least one second before another ping - remove the ping timer
    // code if you do not need this functionality...
    If PingState = SM_FINISHED And (TickGet() - PingTimer) > 1000 Then
        PingState = SM_HOME
        ICMPEndUsage()
    EndIf
End Function

{
****************************************************************************
* Name    : OnData                                                         *
* Purpose : The TCPIP stack is a co-operative state machine. Therefore, we *
*         : cannot wait around for USART data and use a buffered interrupt *
*         : instead. When a CR is received, set a ReadyToRead flag to true *
****************************************************************************
}
Dim ReadyToRead As Boolean
Event OnData()
    // echo char
    usart.writebyte(DataChar)
    If ISRRX.DataChar = #13 Then
        ReadyToRead = true
        usart.writebyte(#10)     // add a linefeed to the echoed data
    EndIf
End Event

//
// main program variables...
//
Dim Result As SHORT
Dim StrIP As String(80)

// program start...
usart.SetBaudrate(br115200)
usart.write(13, 10, "**ICMP test - enter IP addr to ping**", 13, 10)

Init()
#if NET_DHCP
usart.Write("Waiting for DHCP...",13,10)
#else
DisplayConfig
#endif

// use interrupt buffered USART...
ReadyToRead = false
usart.ReadTerminator = 13
ISRRX.Initialize(OnData)

// loop forever...
While true
    // This task performs normal stack task including checking for incoming 
    // packet, type of packet and calling appropriate stack entity to process it.
    NETApp.Task

    // If the ready to read flag has been set, read a string from the interrupt
    // buffer and start the ping...
    If ReadyToRead Then
        ReadyToRead = false
        ISRRX.ReadStr(StrIP,13)
        usart.Write("Ping...")
        Ping(StrIP)
    EndIf

    // Process the ping task state machine...
    Result = PingTask()
    If Result >= 0 Then
        usart.Write(DecToStr(Result), " ms", 13,10)
    ElseIf Result = -1 Then
        usart.Write("timed out",13,10)
    EndIf

    // If DHCP is enabled, display configuration when a DHCP event has occurred...
    #if NET_DHCP
    If NETApp.DHCPEvent() Then
        DisplayConfig()
    EndIf
    #endif
Wend

end program
