//
//-----------------------------------------------------------------------------
// LJCV Electronics eIP-10 board support package
//-----------------------------------------------------------------------------
//
//  18F2620 @ 10/40MHz (64k flash, 3968 ram, 1k eeprom)
//  ENC28J60 10-baseT ethernet
//  25LC256 SPI EEPROM
//  TTL uart (Pickit2 compatable, 38400 max)
//
module user_init

#define EIP10_2620

// call userInit at startup (true/false)
#option USER_AUTO_INIT = true

//
//-----------------------------------------------------------------------------
// config settings
//-----------------------------------------------------------------------------
//
// config
#if (_clock = 10)       // 10MHz xtal
  config OSC = HS
#elseif (_clock = 40)   // 10MHz xtal + PLL
  config OSC = HSPLL
#elseif (_clock = 8)    // intosc
  config  OSC = INTIO67      // Internal oscillator block, port function on RA6 and RA7
#elseif (_clock = 32)   // intosc + pll
  config  OSC = INTIO67      // Internal oscillator block, port function on RA6 and RA7
#else
  #error "unsupported _clock"
#endif

config
    FCMEN = OFF,        // Fail-Safe Clock Monitor disabled
    IESO = OFF,         // Oscillator Switchover mode disabled
    PWRT = OFF,         // PWRT disabled
    BOREN = SBORDIS,    // Brown-out Reset enabled in hardware only (SBOREN is disabled)
    BORV = 3,           // Minimum setting
    WDT = OFF,          // WDT disabled (control is placed on the SWDTEN bit)
    WDTPS = 32768,      // 1:32768
    CCP2MX = PORTC,     // CCP2 input/output is multiplexed with RC1
    PBADEN = ON,        // PORTB<4:0> pins are configured as digital I/O on Reset
    LPT1OSC = OFF,      // Timer1 configured for higher power operation
    MCLRE = ON,         // MCLR pin enabled; RE3 input pin disabled
    STVREN = ON,        // Stack full/underflow will cause Reset
    LVP = OFF,          // Single-Supply ICSP disabled
    XINST = OFF,        // Instruction set extension and Indexed Addressing mode disabled (Legacy mode)
    DEBUG = OFF,        // Background debugger disabled, RB6 and RB7 configured as general purpose I/O pins
    CP0 = OFF,          // Block 0 (000800-003FFFh) not code-protected
    CP1 = OFF,          // Block 1 (004000-007FFFh) not code-protected
    CP2 = OFF,          // Block 2 (008000-00BFFFh) not code-protected
    CP3 = OFF,          // Block 3 (00C000-00FFFFh) not code-protected
    CPB = OFF,          // Boot block (000000-0007FFh) not code-protected
    CPD = OFF,          // Data EEPROM not code-protected
    WRT0 = OFF,         // Block 0 (000800-003FFFh) not write-protected
    WRT1 = OFF,         // Block 1 (004000-007FFFh) not write-protected
    WRT2 = OFF,         // Block 2 (008000-00BFFFh) not write-protected
    WRT3 = OFF,         // Block 3 (00C000-00FFFFh) not write-protected
    WRTC = OFF,         // Configuration registers (300000-3000FFh) not write-protected
    WRTB = OFF,         // Boot Block (000000-0007FFh) not write-protected
    WRTD = OFF,         // Data EEPROM not write-protected
    EBTR0 = OFF,        // Block 0 (000800-003FFFh) not protected from table reads executed in other blocks
    EBTR1 = OFF,        // Block 1 (004000-007FFFh) not protected from table reads executed in other blocks
    EBTR2 = OFF,        // Block 2 (008000-00BFFFh) not protected from table reads executed in other blocks
    EBTR3 = OFF,        // Block 3 (00C000-00FFFFh) not protected from table reads executed in other blocks
    EBTRB = OFF         // Boot Block (000000-0007FFh) not protected from table reads executed in other blocks

// USART support
// note: be sure to use the modified USART.BAS v1.6 module
#option USART_BRGH = true
#option USART_BRG16 = true

//
//-----------------------------------------------------------------------------
// eIP hardware
//-----------------------------------------------------------------------------
//
// LEDs (active high)
public dim 
    LED1 as PORTA.3,
    LED2 as PORTA.4,
    LED3 as PORTA.5

// Switches
// external connector J1-2 (no pullup)
public dim 
    INPUT_SW     as PORTB.0

public dim
    ENC_CS          as PORTC.0,     // ENC28J60 CS
    SPI_EEPROM_CS   as PORTC.1,     // 25LC256 SPI eeprom CS
    _unused_rc2     as PORTC.2,
    SPI_SCK         as PORTC.3,     // spi mssp1
    SPI_MISO        as PORTC.4,
    SPI_MOSI        as PORTC.5,
    TXD             as PORTC.6,     // uart1 TX out (J1-6)
    RXD             as PORTC.7      // uart1 RX in (J1-8)

// ENC interface IO options
#option ENC_CS = PORTC.0        // enc28J60 CS
'#option ENC_RST = PORTA.0      // enc28J60 RST unconnected (default)
'#option ENC_INT = PORTB.1      // enc28J60 INT unconnected (default)
'#option ENC_INT = PORTB.1      // enc28J60 INT U2-4 jumpered to 2620 RB1 U1-22 (J1-1)

//
//-----------------------------------------------------------------------------
// userInitClock() - setup 18F2620 clock
//-----------------------------------------------------------------------------
//
public sub userInitClock()
    const PLLEN = 6     // OSCTUNE PLL bit
    
  #if (_clock = 10) or (_clock = 40)         // HS or HSPLL
    // nothing to do...
  #elseif (_clock = 32)     // INTOSC 8MHZ + 4x PLL
    OSCCON = $70            // INTOSC 8MHz, primary SCS
    OSCTUNE.bits(PLLEN) = 1     // 4x pllen on
  #elseif (_clock = 8)
    OSCCON = $70            // INTOSC 8MHz, primary SCS
    OSCTUNE.bits(PLLEN) = 0     // pllen off
  #else
    #error "unsupported _clock selection"
  #endif

    // allow time for the PLL to lock (if it's enabled)
    if (OSCTUNE.bits(PLLEN) = 1) then
        delayms(2)
    endif
end sub

//
//-----------------------------------------------------------------------------
// userInit() - setup clock, IO pins and other req'd regs
//-----------------------------------------------------------------------------
//
public sub userInit()
    userInitClock()
    
    // setup IO pins
    PORTC = %01000011
    TRISC = %10010000

    ADCON1 = $0F            // all digital IO
    INTCON2.bits(7) = 0     // RBPU enable PORTB pullups (for INPUT_SW)
    
    low(LED1)
    low(LED2)
    low(LED3)

    input(INPUT_SW)
end sub

//
//-----------------------------------------------------------------------------
// userClockCheck() - toggle led at 1ms rate to verify clock setting
//-----------------------------------------------------------------------------
//
public sub userClockCheck()
    while(true)
        LED1 = 1
        delayus(500)
        LED1 = 0
        delayus(500)
    end while
end sub

//
//-----------------------------------------------------------------------------
// module startup
//-----------------------------------------------------------------------------
//
#if (USER_AUTO_INIT)
  userInit()
#else
  #warning "userInit() NOT called at startup"
#endif

end module
